import copy
import os

import numpy as np
import pandas as pd
import pytest

from darts import TimeSeries, concatenate
from darts.dataprocessing.transformers import BoxCox, Scaler
from darts.timeseries import (
    DEFAULT_GLOBAL_STATIC_COV_NAME,
    METADATA_TAG,
    STATIC_COV_TAG,
)
from darts.utils.timeseries_generation import linear_timeseries
from darts.utils.utils import generate_index


def setup_test_case():
    n_groups = 5
    len_ts = 10
    times = (
        pd.concat(
            [
                pd.DataFrame(
                    generate_index(start=pd.Timestamp(2010, 1, 1), length=len_ts)
                )
            ]
            * n_groups,
            axis=0,
        )
        .reset_index(drop=True)
        .rename(columns={0: "times"})
    )

    x = pd.DataFrame(np.random.randn(n_groups * len_ts, 3), columns=["a", "b", "c"])
    static_multivar = pd.DataFrame(
        [
            [i, 0 if j < (len_ts // 2) else 1]
            for i in range(n_groups)
            for j in range(len_ts)
        ],
        columns=["st1", "st2"],
    )

    df_long_multi = pd.DataFrame(
        pd.concat([times, x, static_multivar], axis=1),
    )
    df_long_multi.loc[:, "constant"] = 1
    df_long_uni = df_long_multi.drop(columns=["st2"])
    return n_groups, len_ts, df_long_uni, df_long_multi


def setup_tag(tag, ts):
    if tag == METADATA_TAG:
        x = {"st1": 0.0, "st2": 1.0}
        ts = ts.with_metadata(x)
    else:
        x = pd.Series([0.0, 1.0], index=["st1", "st2"])
        ts = ts.with_static_covariates(x)
    return ts, x


class TestTimeSeriesStaticCovariate:
    n_groups, len_ts, df_long_uni, df_long_multi = setup_test_case()

    @pytest.mark.parametrize("tag", [STATIC_COV_TAG, METADATA_TAG])
    def test_ts_from_x(self, tag, tmpdir_module):
        ts = linear_timeseries(length=10)
        ts, x = setup_tag(tag, ts)
        kwargs = {tag: x}

        self.helper_test_transfer(tag, ts, TimeSeries.from_xarray(ts.data_array()))
        self.helper_test_transfer(
            tag, ts, TimeSeries.from_dataframe(ts.to_dataframe(), **kwargs)
        )
        self.helper_test_transfer(
            tag, ts, TimeSeries.from_series(ts.to_series(), **kwargs)
        )
        self.helper_test_transfer(
            tag,
            ts,
            TimeSeries.from_times_and_values(
                times=ts.time_index,
                values=ts.all_values(),
                columns=ts.components,
                **kwargs,
            ),
        )

        self.helper_test_transfer(
            tag,
            ts,
            TimeSeries.from_values(
                values=ts.all_values(),
                columns=ts.components,
                **kwargs,
            ),
        )

        f_csv = os.path.join(tmpdir_module, "temp_ts.csv")
        f_pkl = os.path.join(tmpdir_module, "temp_ts.pkl")
        ts.to_csv(f_csv)
        ts.to_pickle(f_pkl)
        ts_json = ts.to_json()

        self.helper_test_transfer(
            tag, ts, TimeSeries.from_csv(f_csv, time_col="time", **kwargs)
        )
        self.helper_test_transfer(tag, ts, TimeSeries.from_pickle(f_pkl))
        self.helper_test_transfer(tag, ts, TimeSeries.from_json(ts_json, **kwargs))

    def test_invalid_metadata(self):
        ts = linear_timeseries(length=10)
        with pytest.raises(ValueError) as exc:
            _ = ts.with_metadata(0.0)
        assert (
            str(exc.value)
            == "`metadata` must be of type `dict` mapping metadata attributes to their values."
        )

    @pytest.mark.parametrize("index_type", ["int", "dt", "str"])
    def test_from_group_dataframe(self, index_type):
        """Tests correct extract of TimeSeries groups from a long DataFrame with unsorted (time/integer) index"""
        group = ["a", "a", "a", "b", "b", "b"]
        values = np.arange(len(group))

        if index_type == "int":
            index_expected = pd.RangeIndex(3)
            time = [2, 1, 0, 0, 1, 2]
        else:
            index_expected = pd.date_range("2024-01-01", periods=3)
            time = index_expected[::-1].append(index_expected)
            if index_type == "str":
                time = time.astype(str)

        # create a df with unsorted time
        df = pd.DataFrame({
            "group": group,
            "time": time,
            "x": values,
        })
        ts = TimeSeries.from_group_dataframe(df, group_cols="group", time_col="time")

        # check the time index
        assert ts[0].time_index.equals(index_expected)
        assert ts[1].time_index.equals(index_expected)

        # check the values
        assert (ts[0].values().flatten() == [values[2], values[1], values[0]]).all()
        assert (ts[1].values().flatten() == [values[3], values[4], values[5]]).all()

    def test_timeseries_from_longitudinal_df(self):
        # univariate static covs: only group by "st1", keep static covs "st1"
        value_cols = ["a", "b", "c"]
        ts_groups1 = TimeSeries.from_group_dataframe(
            df=self.df_long_uni,
            group_cols="st1",
            static_cols=None,
            time_col="times",
            value_cols=value_cols,
            metadata_cols=["st1", "constant"],
        )
        assert len(ts_groups1) == self.n_groups
        for i, ts in enumerate(ts_groups1):
            assert ts.static_covariates.index.equals(
                pd.Index([DEFAULT_GLOBAL_STATIC_COV_NAME])
            )
            assert ts.static_covariates.shape == (1, 1)
            assert ts.static_covariates.columns.equals(pd.Index(["st1"]))
            assert (ts.static_covariates_values(copy=False) == [[i]]).all()
            assert ts.metadata == {"st1": i, "constant": 1}

        # multivariate static covs: only group by "st1", keep static covs "st1", "constant"
        ts_groups2 = TimeSeries.from_group_dataframe(
            df=self.df_long_multi,
            group_cols=["st1"],
            static_cols="constant",
            time_col="times",
            value_cols=value_cols,
            metadata_cols=["st1", "constant"],
        )
        assert len(ts_groups2) == self.n_groups
        for i, ts in enumerate(ts_groups2):
            assert ts.static_covariates.shape == (1, 2)
            assert ts.static_covariates.columns.equals(pd.Index(["st1", "constant"]))
            assert (ts.static_covariates_values(copy=False) == [[i, 1]]).all()
            assert ts.metadata == {"st1": i, "constant": 1}

        # multivariate static covs: group by "st1" and "st2", keep static covs "st1", "st2", "constant"
        ts_groups3 = TimeSeries.from_group_dataframe(
            df=self.df_long_multi,
            group_cols=["st1", "st2"],
            static_cols=["constant"],
            time_col="times",
            value_cols=value_cols,
            metadata_cols=["st1", "st2", "constant"],
        )
        assert len(ts_groups3) == self.n_groups * 2
        for idx, ts in enumerate(ts_groups3):
            i = idx // 2
            j = idx % 2
            assert ts.static_covariates.shape == (1, 3)
            assert ts.static_covariates.columns.equals(
                pd.Index(["st1", "st2", "constant"])
            )
            assert (ts.static_covariates_values(copy=False) == [[i, j, 1]]).all()
            assert ts.metadata == {"st1": i, "st2": j, "constant": 1}

        # drop group columns gives same time series with dropped static covariates
        # drop first column
        ts_groups4 = TimeSeries.from_group_dataframe(
            df=self.df_long_multi,
            group_cols=["st1", "st2"],
            static_cols=["constant"],
            time_col="times",
            value_cols=value_cols,
            drop_group_cols=["st1"],
            metadata_cols=["st1", "st2", "constant"],
        )
        assert len(ts_groups4) == self.n_groups * 2
        for idx, ts in enumerate(ts_groups4):
            i = idx // 2
            j = idx % 2
            assert ts.static_covariates.shape == (1, 2)
            assert ts.static_covariates.columns.equals(pd.Index(["st2", "constant"]))
            assert (ts.static_covariates_values(copy=False) == [[j, 1]]).all()
            assert ts.metadata == {"st1": i, "st2": j, "constant": 1}

        # drop last column
        ts_groups5 = TimeSeries.from_group_dataframe(
            df=self.df_long_multi,
            group_cols=["st1", "st2"],
            static_cols=["constant"],
            time_col="times",
            value_cols=value_cols,
            drop_group_cols=["st2"],
            metadata_cols=["st1", "st2", "constant"],
        )
        assert len(ts_groups5) == self.n_groups * 2
        for idx, ts in enumerate(ts_groups5):
            i = idx // 2
            j = idx % 2
            assert ts.static_covariates.shape == (1, 2)
            assert ts.static_covariates.columns.equals(pd.Index(["st1", "constant"]))
            assert (ts.static_covariates_values(copy=False) == [[i, 1]]).all()
            assert ts.metadata == {"st1": i, "st2": j, "constant": 1}

        # drop all columns
        ts_groups6 = TimeSeries.from_group_dataframe(
            df=self.df_long_multi,
            group_cols=["st1", "st2"],
            static_cols=["constant"],
            time_col="times",
            value_cols=value_cols,
            drop_group_cols=["st1", "st2"],
            metadata_cols="constant",
        )
        assert len(ts_groups6) == self.n_groups * 2
        for ts in ts_groups6:
            assert ts.static_covariates.shape == (1, 1)
            assert ts.static_covariates.columns.equals(pd.Index(["constant"]))
            assert (ts.static_covariates_values(copy=False) == [[1]]).all()
            assert ts.metadata == {"constant": 1}

        # drop all static covariates (no `static_cols`, all `group_cols` dropped) and no metadata cols
        ts_groups7 = TimeSeries.from_group_dataframe(
            df=self.df_long_multi,
            group_cols=["st1", "st2"],
            time_col="times",
            value_cols=value_cols,
            drop_group_cols=["st1", "st2"],
        )
        assert len(ts_groups7) == self.n_groups * 2
        for ts in ts_groups7:
            assert ts.static_covariates is None
            assert ts.metadata is None

        ts_groups7_parallel = TimeSeries.from_group_dataframe(
            df=self.df_long_multi,
            group_cols=["st1", "st2"],
            time_col="times",
            value_cols=value_cols,
            drop_group_cols=["st1", "st2"],
            n_jobs=-1,
        )
        assert ts_groups7_parallel == ts_groups7

    def test_from_group_dataframe_invalid_drop_cols(self):
        # drop col is not part of `group_cols`
        with pytest.raises(ValueError) as err:
            _ = TimeSeries.from_group_dataframe(
                df=self.df_long_multi,
                group_cols=["st1"],
                time_col="times",
                value_cols="a",
                drop_group_cols=["invalid"],
            )
        assert str(err.value).endswith("received: {'invalid'}.")

    def test_from_group_dataframe_groups_too_short(self):
        # groups that are too short for TimeSeries requirements should raise an error
        df = copy.deepcopy(self.df_long_multi)
        df.loc[:, "non_static"] = np.arange(len(df))
        with pytest.raises(ValueError) as err:
            _ = TimeSeries.from_group_dataframe(
                df=df,
                group_cols="non_static",
                time_col="times",
                value_cols="a",
            )
        assert str(err.value).startswith(
            "The time index of the provided DataArray is missing the freq attribute"
        )

    def test_from_group_dataframe_not_unique(self):
        # it is assumed that all static / metadata columns have only 1 unique value.
        # it will always pick the first encountered value
        series = TimeSeries.from_group_dataframe(
            df=self.df_long_multi,
            group_cols=["st1"],
            drop_group_cols="st1",
            static_cols="st2",
            time_col="times",
            value_cols="a",
            metadata_cols="st2",
        )
        first_values = [
            df["st2"].values[0] for idx, df in self.df_long_multi.groupby("st1")
        ]
        for s_, val in zip(series, first_values):
            assert s_.static_covariates_values()[0, 0] == val
            assert s_.metadata == {"st2": val}

    def test_with_static_covariates_univariate(self):
        ts = linear_timeseries(length=10)
        static_covs_series = pd.Series([0.0, 1.0], index=["st1", "st2"])
        static_covs_df = pd.DataFrame([[0.0, 1.0]], columns=["st1", "st2"])

        # check immutable
        ts.with_static_covariates(static_covs_series)
        assert not ts.has_static_covariates

        # from Series
        ts = ts.with_static_covariates(static_covs_series)
        assert ts.has_static_covariates
        np.testing.assert_almost_equal(
            ts.static_covariates_values(copy=False),
            np.expand_dims(static_covs_series.values, -1).T,
        )
        assert ts.static_covariates.index.equals(ts.components)

        # from DataFrame
        ts = ts.with_static_covariates(static_covs_df)
        assert ts.has_static_covariates
        np.testing.assert_almost_equal(
            ts.static_covariates_values(copy=False), static_covs_df.values
        )
        assert ts.static_covariates.index.equals(ts.components)

        # with None
        ts = ts.with_static_covariates(None)
        assert ts.static_covariates is None
        assert not ts.has_static_covariates

        # only pd.Series, pd.DataFrame or None
        with pytest.raises(ValueError):
            _ = ts.with_static_covariates([1, 2, 3])

        # multivariate does not work with univariate TimeSeries
        with pytest.raises(ValueError):
            static_covs_multi = pd.concat([static_covs_series] * 2, axis=1).T
            _ = ts.with_static_covariates(static_covs_multi)

    def test_static_covariates_values(self):
        ts = linear_timeseries(length=10)
        static_covs = pd.DataFrame([[0.0, 1.0]], columns=["st1", "st2"])
        ts = ts.with_static_covariates(static_covs)

        # changing values of copy should not change original DataFrame
        vals = ts.static_covariates_values(copy=True)
        vals[:] = -1.0
        assert (ts.static_covariates_values(copy=False) != -1.0).all()

        # changing values of view should change original DataFrame
        vals = ts.static_covariates_values(copy=False)
        vals[:] = -1.0
        assert (ts.static_covariates_values(copy=False) == -1.0).all()

        ts = ts.with_static_covariates(None)
        assert ts.static_covariates is None
        assert ts.static_covariates_values() is None

    def test_metadata_values(self):
        ts = linear_timeseries(length=10)
        assert not ts.has_metadata

        metadata = {"st1": 0, "st2": 1}
        ts = ts.with_metadata(metadata)

        assert ts.has_metadata
        assert ts.metadata == metadata
        # changing values of metadata is allowed
        ts.metadata["st1"] = 2
        assert ts.metadata == {"st1": 2, "st2": 1}

        # removing metadata
        ts = ts.with_metadata(None)
        assert ts.metadata is None

    def test_with_static_covariates_multivariate(self):
        ts = linear_timeseries(length=10)
        ts_multi = ts.stack(ts)
        static_covs = pd.DataFrame([[0.0, 1.0], [0.0, 1.0]], columns=["st1", "st2"])

        # from univariate static covariates
        ts_multi = ts_multi.with_static_covariates(static_covs.loc[0])
        assert ts_multi.static_covariates.index.equals(
            pd.Index([DEFAULT_GLOBAL_STATIC_COV_NAME])
        )
        assert ts_multi.static_covariates.columns.equals(static_covs.columns)
        np.testing.assert_almost_equal(
            ts_multi.static_covariates_values(copy=False), static_covs.loc[0:0].values
        )

        # from multivariate static covariates
        ts_multi = ts_multi.with_static_covariates(static_covs)
        assert ts_multi.static_covariates.index.equals(ts_multi.components)
        assert ts_multi.static_covariates.columns.equals(static_covs.columns)
        np.testing.assert_almost_equal(
            ts_multi.static_covariates_values(copy=False), static_covs.values
        )

        # raise an error if multivariate static covariates columns don't match the number of components in the series
        with pytest.raises(ValueError):
            _ = ts_multi.with_static_covariates(pd.concat([static_covs] * 2, axis=0))

    def test_stack(self):
        ts_uni = linear_timeseries(length=10)
        ts_multi = ts_uni.stack(ts_uni)

        static_covs_uni1 = pd.DataFrame([[0, 1]], columns=["st1", "st2"]).astype(int)
        static_covs_uni2 = pd.DataFrame([[3, 4]], columns=["st3", "st4"]).astype(int)
        static_covs_uni3 = pd.DataFrame(
            [[2, 3, 4]], columns=["st1", "st2", "st3"]
        ).astype(int)

        static_covs_multi = pd.DataFrame(
            [[0, 0], [1, 1]], columns=["st1", "st2"]
        ).astype(int)

        ts_uni = ts_uni.with_static_covariates(static_covs_uni1)
        ts_multi = ts_multi.with_static_covariates(static_covs_multi)

        # valid static covariates for concatenation/stack
        ts_stacked1 = ts_uni.stack(ts_uni)
        assert ts_stacked1.static_covariates.index.equals(ts_stacked1.components)
        np.testing.assert_almost_equal(
            ts_stacked1.static_covariates_values(copy=False),
            pd.concat([ts_uni.static_covariates] * 2, axis=0).values,
        )

        # valid static covariates for concatenation/stack: first only has static covs
        # -> this gives multivar ts with univar static covs
        ts_stacked2 = ts_uni.stack(ts_uni.with_static_covariates(None))
        np.testing.assert_almost_equal(
            ts_stacked2.static_covariates_values(copy=False),
            ts_uni.static_covariates_values(copy=False),
        )

        # mismatch between column names
        with pytest.raises(ValueError):
            _ = ts_uni.stack(ts_uni.with_static_covariates(static_covs_uni2))

        # mismatch between number of covariates
        with pytest.raises(ValueError):
            _ = ts_uni.stack(ts_uni.with_static_covariates(static_covs_uni3))

        # valid univar ts with univar static covariates + multivar ts with multivar static covariates
        ts_stacked3 = ts_uni.stack(ts_multi)
        np.testing.assert_almost_equal(
            ts_stacked3.static_covariates_values(copy=False),
            pd.concat(
                [ts_uni.static_covariates, ts_multi.static_covariates], axis=0
            ).values,
        )

        # invalid univar ts with univar static covariates + multivar ts with univar static covariates
        with pytest.raises(ValueError):
            _ = ts_uni.stack(ts_multi.with_static_covariates(static_covs_uni1))

    def test_stack_metadata(self):
        metadata = {"a": 0, "b": 1}
        ts = linear_timeseries(length=10)
        ts_md = ts.with_metadata(metadata)

        # stacking will always use the metadata of `self`
        assert ts.stack(ts).metadata is None
        assert ts.stack(ts_md).metadata is None
        assert ts_md.stack(ts).metadata == metadata
        assert ts_md.stack(ts_md).metadata == metadata

    def test_concatenate_dim_component(self):
        """
        test concatenation with static covariates along component dimension (axis=1)
        Along component dimension, we concatenate/transfer the static covariates of the series only if one of
        below cases applies:
        1)  concatenate when for each series the number of static cov components is equal to the number of
            components in the series. The static variable names (columns in series.static_covariates) must be
            identical across all series
        2)  if only the first series contains static covariates transfer only those
        3)  if `ignore_static_covarites=True`, case 1) is ignored and only the static covariates of the first
            series are transferred
        """
        ts_uni = linear_timeseries(length=10)
        ts_multi = ts_uni.stack(ts_uni)

        static_covs_uni1 = pd.DataFrame([[0, 1]], columns=["st1", "st2"]).astype(int)
        static_covs_uni2 = pd.DataFrame([[3, 4]], columns=["st3", "st4"]).astype(int)
        static_covs_uni3 = pd.DataFrame(
            [[2, 3, 4]], columns=["st1", "st2", "st3"]
        ).astype(int)

        static_covs_multi = pd.DataFrame(
            [[0, 0], [1, 1]], columns=["st1", "st2"]
        ).astype(int)

        ts_uni_static_uni1 = ts_uni.with_static_covariates(static_covs_uni1)
        ts_uni_static_uni2 = ts_uni.with_static_covariates(static_covs_uni2)
        ts_uni_static_uni3 = ts_uni.with_static_covariates(static_covs_uni3)

        ts_multi_static_uni1 = ts_multi.with_static_covariates(static_covs_uni1)
        ts_multi_static_multi = ts_multi.with_static_covariates(static_covs_multi)

        # concatenation without covariates
        ts_concat = concatenate([ts_uni, ts_uni], axis=1)
        assert ts_concat.static_covariates is None

        # concatenation along component dimension results in multi component static covariates
        ts_concat = concatenate([ts_uni_static_uni1, ts_uni_static_uni1], axis=1)
        assert ts_concat.static_covariates.shape == (2, 2)
        assert ts_concat.components.equals(ts_concat.static_covariates.index)
        np.testing.assert_almost_equal(
            ts_concat.static_covariates_values(copy=False),
            pd.concat([static_covs_uni1] * 2, axis=0).values,
        )

        # concatenation with inconsistent static variable names should fail ...
        with pytest.raises(ValueError):
            _ = concatenate([ts_uni_static_uni1, ts_uni_static_uni2], axis=1)

        # ... by ignoring the static covariates, it should work and take only the covariates of the first series
        ts_concat = concatenate(
            [ts_uni_static_uni1, ts_uni_static_uni2],
            axis=1,
            ignore_static_covariates=True,
        )
        assert ts_concat.static_covariates.shape == (1, 2)
        assert ts_concat.static_covariates.index.equals(
            pd.Index([DEFAULT_GLOBAL_STATIC_COV_NAME])
        )
        np.testing.assert_almost_equal(
            ts_concat.static_covariates_values(copy=False),
            ts_uni_static_uni1.static_covariates_values(copy=False),
        )

        # concatenation with inconsistent number of static covariates should fail ...
        with pytest.raises(ValueError):
            _ = concatenate([ts_uni_static_uni1, ts_uni_static_uni3], axis=1)

        # concatenation will only work if for each series the number of static cov components is equal to the
        # number of components in the series
        with pytest.raises(ValueError):
            _ = concatenate([ts_uni_static_uni1, ts_multi_static_uni1], axis=1)

        ts_concat = concatenate([ts_uni_static_uni1, ts_multi_static_multi], axis=1)
        assert ts_concat.static_covariates.shape == (ts_concat.n_components, 2)
        assert ts_concat.components.equals(ts_concat.static_covariates.index)
        np.testing.assert_almost_equal(
            ts_concat.static_covariates_values(copy=False),
            pd.concat([static_covs_uni1, static_covs_multi], axis=0),
        )

    def test_concatenate_dim_time(self):
        """
        Test concatenation with static covariates along time dimension (axis=0)
        Along time dimension, we only take the static covariates of the first series (as static covariates are
        time-independant).
        """
        static_covs_left = pd.DataFrame([[0, 1]], columns=["st1", "st2"]).astype(int)
        static_covs_right = pd.DataFrame([[3, 4]], columns=["st3", "st4"]).astype(int)

        ts_left = linear_timeseries(length=10).with_static_covariates(static_covs_left)
        ts_right = linear_timeseries(
            length=10, start=ts_left.end_time() + ts_left.freq
        ).with_static_covariates(static_covs_right)

        ts_concat = concatenate([ts_left, ts_right], axis=0)
        assert ts_concat.static_covariates.equals(ts_left.static_covariates)

    def test_concatenate_dim_samples(self):
        """
        Test concatenation with static covariates along sample dimension (axis=2)
        Along sample dimension, we only take the static covariates of the first series (as we components and
        time don't change).
        """
        static_covs_left = pd.DataFrame([[0, 1]], columns=["st1", "st2"]).astype(int)
        static_covs_right = pd.DataFrame([[3, 4]], columns=["st3", "st4"]).astype(int)

        ts_left = linear_timeseries(length=10).with_static_covariates(static_covs_left)
        ts_right = linear_timeseries(length=10).with_static_covariates(
            static_covs_right
        )

        ts_concat = concatenate([ts_left, ts_right], axis=2)
        assert ts_concat.static_covariates.equals(ts_left.static_covariates)

    def test_concatenate_metadata(self):
        metadata = {"a": 0, "b": 1}
        ts = linear_timeseries(length=10)
        ts_md = ts.with_metadata(metadata)

        # concatenate will always use the metadata of `self`
        for axis in [1, 2]:
            assert ts.concatenate(ts, axis=axis).metadata is None
            assert ts.concatenate(ts_md, axis=axis).metadata is None
            assert ts_md.concatenate(ts, axis=axis).metadata == metadata
            assert ts_md.concatenate(ts_md, axis=axis).metadata == metadata

        # over axis=0 requires shifting
        assert ts.concatenate(ts_md.shift(10), axis=0).metadata is None
        assert ts_md.concatenate(ts.shift(10), axis=0).metadata == metadata

    def test_scalers_with_static_covariates(self):
        ts = linear_timeseries(start_value=1.0, end_value=2.0, length=10)
        static_covs = pd.Series([0.0, 2.0], index=["st1", "st2"])
        ts = ts.with_static_covariates(static_covs)

        for scaler_cls in [Scaler, BoxCox]:
            scaler = scaler_cls()
            ts_scaled = scaler.fit_transform(ts)
            assert ts_scaled.static_covariates.equals(ts.static_covariates)

            ts_inv = scaler.inverse_transform(ts_scaled)
            assert ts_inv.static_covariates.equals(ts.static_covariates)

    def test_non_numerical_static_covariates(self):
        static_covs = pd.DataFrame([["a", 0], ["b", 1]], columns=["cat", "num"])
        assert static_covs.dtypes["num"] == "int64"

        ts = TimeSeries.from_values(
            values=np.random.random((10, 2))
        ).with_static_covariates(static_covs)
        assert ts.static_covariates.dtypes["num"] == ts.dtype == "float64"
        assert isinstance(ts.static_covariates.dtypes["cat"], object)

        ts = ts.astype(np.float32)
        assert ts.static_covariates.dtypes["num"] == ts.dtype == "float32"
        assert isinstance(ts.static_covariates.dtypes["cat"], object)

    def test_non_numerical_metadata(self):
        metadata = {"a": 0, "b": "foo"}
        ts = linear_timeseries(length=10)
        assert ts.with_metadata(metadata).metadata == metadata

    def test_get_item(self):
        # multi component static covariates
        static_covs = pd.DataFrame([["a", 0], ["b", 1]], columns=["cat", "num"])
        ts = TimeSeries.from_values(
            values=np.random.random((10, 2)), columns=["comp1", "comp2"]
        ).with_static_covariates(static_covs)

        assert ts.static_covariates.index.equals(ts.components)

        ts0 = ts[0]
        assert ts0.static_covariates.index.equals(ts.components)
        assert isinstance(ts0.static_covariates, pd.DataFrame)
        ts1 = ts["comp1"]
        assert ts1.static_covariates.index.equals(pd.Index(["comp1"]))
        assert isinstance(ts1.static_covariates, pd.DataFrame)
        ts2 = ts["comp2"]
        assert ts2.static_covariates.index.equals(pd.Index(["comp2"]))
        assert isinstance(ts2.static_covariates, pd.DataFrame)
        ts3 = ts["comp1":"comp2"]
        assert ts3.static_covariates.index.equals(pd.Index(["comp1", "comp2"]))
        assert isinstance(ts3.static_covariates, pd.DataFrame)
        ts4 = ts[["comp1", "comp2"]]
        assert ts4.static_covariates.index.equals(pd.Index(["comp1", "comp2"]))
        assert isinstance(ts4.static_covariates, pd.DataFrame)

        # uni/global component static covariates
        static_covs = pd.DataFrame([["a", 0]], columns=["cat", "num"])
        ts = TimeSeries.from_values(
            values=np.random.random((10, 3)), columns=["comp1", "comp2", "comp3"]
        ).with_static_covariates(static_covs)

        # 1) when static covs have 1 component but series is multivariate -> static covariate component name is set to
        # "global_components"
        assert ts.static_covariates.index.equals(
            pd.Index([DEFAULT_GLOBAL_STATIC_COV_NAME])
        )
        ts0 = ts[0]
        assert ts0.static_covariates.index.equals(
            pd.Index([DEFAULT_GLOBAL_STATIC_COV_NAME])
        )
        assert isinstance(ts0.static_covariates, pd.DataFrame)
        ts1 = ts["comp1":"comp3"]
        assert ts1.static_covariates.index.equals(
            pd.Index([DEFAULT_GLOBAL_STATIC_COV_NAME])
        )
        assert isinstance(ts1.static_covariates, pd.DataFrame)
        ts2 = ts[["comp1", "comp2", "comp3"]]
        assert ts2.static_covariates.index.equals(
            pd.Index([DEFAULT_GLOBAL_STATIC_COV_NAME])
        )
        assert isinstance(ts2.static_covariates, pd.DataFrame)

        # 2) if number of static cov components match the number of components in the series -> static covariate
        # component names are set to be equal to series component names
        ts3 = ts["comp1"]
        assert ts3.static_covariates.index.equals(pd.Index(["comp1"]))
        assert isinstance(ts3.static_covariates, pd.DataFrame)
        ts4 = ts["comp2"]
        assert ts4.static_covariates.index.equals(pd.Index(["comp2"]))
        assert isinstance(ts4.static_covariates, pd.DataFrame)

    @pytest.mark.parametrize(
        "get_item",
        [0, slice(0, 2), "comp1", slice("comp1", "comp2"), ["comp1", "comp2"]],
    )
    def test_get_item_metadata(self, get_item):
        metadata = {"a": 0, "b": 1}
        ts = TimeSeries.from_values(
            values=np.random.random((10, 2)), columns=["comp1", "comp2"]
        ).with_metadata(metadata)
        assert ts[get_item].metadata == metadata

    @pytest.mark.parametrize("tag", [STATIC_COV_TAG, METADATA_TAG])
    def test_operations(self, tag):
        ts = TimeSeries.from_values(values=np.random.random((10, 2)))
        ts, x = setup_tag(tag, ts)

        # arithmetics with series (left) and non-series (right)
        self.helper_test_transfer(tag, ts, ts / 3)
        self.helper_test_transfer(tag, ts, ts * 3)
        self.helper_test_transfer(tag, ts, ts**3)
        self.helper_test_transfer(tag, ts, ts + 3)
        self.helper_test_transfer(tag, ts, ts - 3)

        # conditions
        self.helper_test_transfer_xa(tag, ts, ts < 3)
        self.helper_test_transfer_xa(tag, ts, ts >= 3)
        self.helper_test_transfer_xa(tag, ts, ts > 3)
        self.helper_test_transfer_xa(tag, ts, ts >= 3)

        # arithmetics with non-series (left) and series (right)
        self.helper_test_transfer(tag, ts, 3 * ts)
        self.helper_test_transfer(tag, ts, 3 + ts)
        self.helper_test_transfer(tag, ts, 3 - ts)
        # conditions
        self.helper_test_transfer_xa(tag, ts, 3 > ts)
        self.helper_test_transfer_xa(tag, ts, 3 >= ts)
        self.helper_test_transfer_xa(tag, ts, 3 < ts)
        self.helper_test_transfer_xa(tag, ts, 3 <= ts)

        # arithmetics with two series
        self.helper_test_transfer(tag, ts, ts / ts)
        self.helper_test_transfer(tag, ts, ts * ts)
        self.helper_test_transfer(tag, ts, ts**ts)
        self.helper_test_transfer(tag, ts, ts + ts)
        self.helper_test_transfer(tag, ts, ts - ts)
        # conditions
        self.helper_test_transfer_xa(tag, ts, ts > ts)
        self.helper_test_transfer_xa(tag, ts, ts >= ts)
        self.helper_test_transfer_xa(tag, ts, ts < ts)
        self.helper_test_transfer_xa(tag, ts, ts <= ts)

        # other operations
        self.helper_test_transfer(tag, ts, abs(ts))
        self.helper_test_transfer(tag, ts, -ts)
        self.helper_test_transfer(tag, ts, round(ts, 2))

    @pytest.mark.parametrize("tag", [STATIC_COV_TAG, METADATA_TAG])
    def test_ts_methods(self, tag):
        ts = linear_timeseries(length=10, start_value=1.0, end_value=2.0).astype(
            "float64"
        )
        ts, x = setup_tag(tag, ts)
        kwargs = {tag: x}

        ts_stoch = ts.from_times_and_values(
            times=ts.time_index,
            values=np.random.randint(low=0, high=10, size=(10, 1, 3)),
            **kwargs,
        )

        if tag == STATIC_COV_TAG:
            assert ts.static_covariates.dtypes.iloc[0] == "float64"
            ts = ts.astype("float32")
            assert ts.static_covariates.dtypes.iloc[0] == "float32"
            assert ts_stoch.static_covariates.index.equals(ts_stoch.components)

        self.helper_test_transfer(tag, ts, ts.with_values(ts.all_values()))
        self.helper_test_transfer(
            tag,
            ts,
            ts.with_columns_renamed(ts.components.tolist(), ts.components.tolist()),
        )
        self.helper_test_transfer(tag, ts, ts.copy())
        self.helper_test_transfer(tag, ts, ts.mean())
        self.helper_test_transfer(tag, ts, ts.median())
        self.helper_test_transfer(tag, ts, ts.sum())
        self.helper_test_transfer(tag, ts, ts.min())
        self.helper_test_transfer(tag, ts, ts.max())
        self.helper_test_transfer(tag, ts, ts.head())
        self.helper_test_transfer(tag, ts, ts.tail())
        self.helper_test_transfer(tag, ts, ts.split_after(0.5)[0])
        self.helper_test_transfer(tag, ts, ts.split_after(0.5)[1])
        self.helper_test_transfer(tag, ts, ts.split_before(0.5)[0])
        self.helper_test_transfer(tag, ts, ts.split_before(0.5)[1])
        self.helper_test_transfer(tag, ts, ts.drop_before(0.5))
        self.helper_test_transfer(tag, ts, ts.drop_after(0.5))
        self.helper_test_transfer(
            tag, ts, ts.slice(ts.start_time() + ts.freq, ts.end_time() - ts.freq)
        )
        self.helper_test_transfer(tag, ts, ts.slice_n_points_after(ts.start_time(), 5))
        self.helper_test_transfer(tag, ts, ts.slice_n_points_before(ts.end_time(), 5))
        self.helper_test_transfer(tag, ts, ts.slice_intersect(ts[2:]))
        self.helper_test_transfer(tag, ts, ts.strip())
        self.helper_test_transfer(tag, ts, ts.longest_contiguous_slice())
        self.helper_test_transfer(tag, ts, ts.rescale_with_value(2.0))
        self.helper_test_transfer(tag, ts, ts.shift(2.0))
        self.helper_test_transfer(tag, ts, ts.diff())
        self.helper_test_transfer(tag, ts, ts.univariate_component(0))
        self.helper_test_transfer(tag, ts, ts.map(lambda x: x + 1))
        self.helper_test_transfer(tag, ts, ts.resample(ts.freq))
        self.helper_test_transfer(tag, ts, ts[:5].append(ts[5:]))
        self.helper_test_transfer(tag, ts, ts.append_values(ts.all_values()))

        self.helper_test_transfer(tag, ts_stoch, ts_stoch.var())
        self.helper_test_transfer(tag, ts_stoch, ts_stoch.std())
        self.helper_test_transfer(tag, ts_stoch, ts_stoch.skew())
        self.helper_test_transfer(tag, ts_stoch, ts_stoch.kurtosis())

        # will append "_quantile" to component names
        self.helper_test_transfer_values(tag, ts_stoch, ts_stoch.quantile_timeseries())
        self.helper_test_transfer_values(tag, ts_stoch, ts_stoch.quantile(0.5))
        # will change component names
        self.helper_test_transfer_values(tag, ts, ts.add_datetime_attribute("hour"))
        self.helper_test_transfer_values(tag, ts, ts.add_holidays("US"))

    @staticmethod
    def helper_test_transfer(tag, ts, ts_new):
        """static cov or metadata must be identical"""
        if tag == STATIC_COV_TAG:
            assert ts_new.static_covariates.equals(ts.static_covariates)
        else:  # metadata
            assert ts_new.metadata == ts.metadata

    @staticmethod
    def helper_test_transfer_xa(tag, ts, xa_new):
        """static cov or metadata must be identical between xarray and TimeSeries"""
        if tag == STATIC_COV_TAG:
            assert xa_new.attrs[STATIC_COV_TAG].equals(ts.static_covariates)
        else:  # metadata
            assert xa_new.attrs[METADATA_TAG] == ts.metadata

    @staticmethod
    def helper_test_transfer_values(tag, ts, ts_new):
        """values of static cov or metadata must match but not row index (component names).
        I.e. series.quantile_timeseries() adds "_quantiles" to component names
        """
        if tag == STATIC_COV_TAG:
            assert not ts_new.static_covariates.index.equals(ts.components)
            np.testing.assert_almost_equal(
                ts_new.static_covariates_values(copy=False),
                ts.static_covariates_values(copy=False),
            )
        else:  # metadata
            assert ts_new.metadata == ts.metadata



"""
Transformer Model
-----------------
"""

import math
from typing import Optional, Union

import torch
import torch.nn as nn

from darts.logging import get_logger, raise_if, raise_if_not, raise_log
from darts.models.components import glu_variants, layer_norm_variants
from darts.models.components.glu_variants import GLU_FFN
from darts.models.components.transformer import (
    CustomFeedForwardDecoderLayer,
    CustomFeedForwardEncoderLayer,
)
from darts.models.forecasting.pl_forecasting_module import (
    PLPastCovariatesModule,
    io_processor,
)
from darts.models.forecasting.torch_forecasting_model import PastCovariatesTorchModel
from darts.utils.torch import MonteCarloDropout

logger = get_logger(__name__)


BUILT_IN = ["relu", "gelu"]
FFN = GLU_FFN + BUILT_IN


def _generate_coder(
    d_model,
    dim_ff,
    dropout,
    nhead,
    num_layers,
    norm_layer,
    coder_cls,
    layer_cls,
    ffn_cls,
):
    """Generates an Encoder or Decoder with one of Darts' Feed-forward Network variants.
    Parameters
    ----------
    coder_cls
        Either `torch.nn.TransformerEncoder` or `...TransformerDecoder`
    layer_cls
        Either `darts.models.components.transformer.CustomFeedForwardEncoderLayer`,
        `...CustomFeedForwardDecoderLayer`, `nn.TransformerEncoderLayer`, or `nn.TransformerDecoderLayer`.
    ffn_cls
        One of Darts' Position-wise Feed-Forward Network variants `from darts.models.components.glu_variants`
    """

    ffn = (
        dict(ffn=ffn_cls(d_model=d_model, d_ff=dim_ff, dropout=dropout))
        if ffn_cls
        else dict()
    )
    layer = layer_cls(
        **ffn,
        dropout=dropout,
        d_model=d_model,
        nhead=nhead,
        dim_feedforward=dim_ff,
    )
    return coder_cls(
        layer,
        num_layers=num_layers,
        norm=norm_layer(d_model),
    )


# This implementation of positional encoding is taken from the PyTorch documentation:
# https://pytorch.org/tutorials/beginner/transformer_tutorial.html
class _PositionalEncoding(nn.Module):
    def __init__(self, d_model, dropout=0.1, max_len=500):
        """An implementation of positional encoding as described in 'Attention is All you Need' by Vaswani et al. (2017)

        Parameters
        ----------
        d_model
            The number of expected features in the transformer encoder/decoder inputs.
            Last dimension of the input.
        dropout
            Fraction of neurons affected by Dropout (default=0.1).
        max_len
            The dimensionality of the computed positional encoding array.
            Only its first "input_size" elements will be considered in the output.

        Inputs
        ------
        x of shape `(batch_size, input_size, d_model)`
            Tensor containing the embedded time series.

        Outputs
        -------
        y of shape `(batch_size, input_size, d_model)`
            Tensor containing the embedded time series enhanced with positional encoding.
        """
        super().__init__()
        self.dropout = MonteCarloDropout(p=dropout)

        pe = torch.zeros(max_len, d_model)
        position = torch.arange(0, max_len, dtype=torch.float).unsqueeze(1)
        div_term = torch.exp(
            torch.arange(0, d_model, 2).float() * (-math.log(10000.0) / d_model)
        )
        pe[:, 0::2] = torch.sin(position * div_term)
        pe[:, 1::2] = torch.cos(position * div_term)
        pe = pe.unsqueeze(0).transpose(0, 1)
        self.register_buffer("pe", pe)

    def forward(self, x):
        x = x + self.pe[: x.size(0), :]
        return self.dropout(x)


class _TransformerModule(PLPastCovariatesModule):
    def __init__(
        self,
        input_size: int,
        output_size: int,
        nr_params: int,
        d_model: int,
        nhead: int,
        num_encoder_layers: int,
        num_decoder_layers: int,
        dim_feedforward: int,
        dropout: float,
        activation: str,
        norm_type: Union[str, nn.Module, None] = None,
        custom_encoder: Optional[nn.Module] = None,
        custom_decoder: Optional[nn.Module] = None,
        **kwargs,
    ):
        """PyTorch module implementing a Transformer to be used in `TransformerModel`.

        PyTorch module implementing a simple encoder-decoder transformer architecture.

        Parameters
        ----------
        input_size
            The dimensionality of the TimeSeries instances that will be fed to the the fit and predict functions.
        output_size
            The dimensionality of the output time series.
        nr_params
            The number of parameters of the likelihood (or 1 if no likelihood is used).
        d_model
            The number of expected features in the transformer encoder/decoder inputs.
        nhead
            The number of heads in the multiheadattention model.
        num_encoder_layers
            The number of encoder layers in the encoder.
        num_decoder_layers
            The number of decoder layers in the decoder.
        dim_feedforward
            The dimension of the feedforward network model.
        dropout
            Fraction of neurons affected by Dropout.
        activation
            The activation function of encoder/decoder intermediate layer.
        norm_type: str | nn.Module | None
            The type of LayerNorm variant to use.
        custom_encoder
            A custom transformer encoder provided by the user (default=None).
        custom_decoder
            A custom transformer decoder provided by the user (default=None).
        **kwargs
            All parameters required for :class:`darts.models.forecasting.pl_forecasting_module.PLForecastingModule`
            base class.

        Inputs
        ------
        x of shape `(batch_size, input_chunk_length, input_size)`
            Tensor containing the features of the input sequence.

        Outputs
        -------
        y of shape `(batch_size, output_chunk_length, target_size, nr_params)`
            Tensor containing the prediction at the last time step of the sequence.
        """

        super().__init__(**kwargs)

        self.input_size = input_size
        self.target_size = output_size
        self.nr_params = nr_params
        self.target_length = self.output_chunk_length

        self.encoder = nn.Linear(input_size, d_model)
        self.positional_encoding = _PositionalEncoding(
            d_model, dropout, self.input_chunk_length
        )

        if isinstance(norm_type, str):
            try:
                self.layer_norm = getattr(layer_norm_variants, norm_type)
            except AttributeError:
                raise_log(
                    AttributeError("please provide a valid layer norm type"),
                )
        else:
            self.layer_norm = norm_type

        raise_if_not(activation in FFN, f"'{activation}' is not in {FFN}")
        if activation in GLU_FFN:
            raise_if(
                custom_encoder is not None or custom_decoder is not None,
                "Cannot use `custom_encoder` or `custom_decoder` along with an `activation` from "
                f"{GLU_FFN}",
                logger=logger,
            )
            # use glu variant feed-forward layers
            ffn_cls = getattr(glu_variants, activation)

            # custom feed-forward layers have activation built-in. reset activation
            activation = None

            custom_encoder = _generate_coder(
                d_model,
                dim_feedforward,
                dropout,
                nhead,
                num_encoder_layers,
                self.layer_norm if self.layer_norm else nn.LayerNorm,
                coder_cls=nn.TransformerEncoder,
                layer_cls=CustomFeedForwardEncoderLayer,
                ffn_cls=ffn_cls,
            )

            custom_decoder = _generate_coder(
                d_model,
                dim_feedforward,
                dropout,
                nhead,
                num_decoder_layers,
                self.layer_norm if self.layer_norm else nn.LayerNorm,
                coder_cls=nn.TransformerDecoder,
                layer_cls=CustomFeedForwardDecoderLayer,
                ffn_cls=ffn_cls,
            )

        # if layer norm set and no GLU variant is used
        if self.layer_norm and custom_decoder is None:
            custom_encoder = _generate_coder(
                d_model,
                dim_feedforward,
                dropout,
                nhead,
                num_encoder_layers,
                self.layer_norm,
                coder_cls=nn.TransformerEncoder,
                layer_cls=nn.TransformerEncoderLayer,
                ffn_cls=None,
            )

            custom_decoder = _generate_coder(
                d_model,
                dim_feedforward,
                dropout,
                nhead,
                num_decoder_layers,
                self.layer_norm,
                coder_cls=nn.TransformerDecoder,
                layer_cls=nn.TransformerDecoderLayer,
                ffn_cls=None,
            )

        # Defining the Transformer module
        self.transformer = nn.Transformer(
            d_model=d_model,
            nhead=nhead,
            num_encoder_layers=num_encoder_layers,
            num_decoder_layers=num_decoder_layers,
            dim_feedforward=dim_feedforward,
            dropout=dropout,
            activation=activation,
            custom_encoder=custom_encoder,
            custom_decoder=custom_decoder,
        )

        self.decoder = nn.Linear(
            d_model, self.target_length * self.target_size * self.nr_params
        )

    def _create_transformer_inputs(self, data):
        # '_TimeSeriesSequentialDataset' stores time series in the
        # (batch_size, input_chunk_length, input_size) format. PyTorch's nn.Transformer
        # module needs it the (input_chunk_length, batch_size, input_size) format.
        # Therefore, the first two dimensions need to be swapped.
        src = data.permute(1, 0, 2)
        tgt = src[-1:, :, :]

        return src, tgt

    @io_processor
    def forward(self, x_in: tuple):
        data, _ = x_in
        # Here we create 'src' and 'tgt', the inputs for the encoder and decoder
        # side of the Transformer architecture
        src, tgt = self._create_transformer_inputs(data)

        # "math.sqrt(self.input_size)" is a normalization factor
        # see section 3.2.1 in 'Attention is All you Need' by Vaswani et al. (2017)
        src = self.encoder(src) * math.sqrt(self.input_size)
        src = self.positional_encoding(src)

        tgt = self.encoder(tgt) * math.sqrt(self.input_size)
        tgt = self.positional_encoding(tgt)

        x = self.transformer(src=src, tgt=tgt)
        out = self.decoder(x)

        # Here we change the data format
        # from (1, batch_size, output_chunk_length * output_size)
        # to (batch_size, output_chunk_length, output_size, nr_params)
        predictions = out[0, :, :]
        predictions = predictions.view(
            -1, self.target_length, self.target_size, self.nr_params
        )

        return predictions


class TransformerModel(PastCovariatesTorchModel):
    def __init__(
        self,
        input_chunk_length: int,
        output_chunk_length: int,
        output_chunk_shift: int = 0,
        d_model: int = 64,
        nhead: int = 4,
        num_encoder_layers: int = 3,
        num_decoder_layers: int = 3,
        dim_feedforward: int = 512,
        dropout: float = 0.1,
        activation: str = "relu",
        norm_type: Union[str, nn.Module, None] = None,
        custom_encoder: Optional[nn.Module] = None,
        custom_decoder: Optional[nn.Module] = None,
        **kwargs,
    ):
        """Transformer model

        Transformer is a state-of-the-art deep learning model introduced in 2017. It is an encoder-decoder
        architecture whose core feature is the 'multi-head attention' mechanism, which is able to
        draw intra-dependencies within the input vector and within the output vector ('self-attention')
        as well as inter-dependencies between input and output vectors ('encoder-decoder attention').
        The multi-head attention mechanism is highly parallelizable, which makes the transformer architecture
        very suitable to be trained with GPUs.

        The transformer architecture implemented here is based on [1]_.

        This model supports past covariates (known for `input_chunk_length` points before prediction time).

        Parameters
        ----------
        input_chunk_length
            Number of time steps in the past to take as a model input (per chunk). Applies to the target
            series, and past and/or future covariates (if the model supports it).
        output_chunk_length
            Number of time steps predicted at once (per chunk) by the internal model. Also, the number of future values
            from future covariates to use as a model input (if the model supports future covariates). It is not the same
            as forecast horizon `n` used in `predict()`, which is the desired number of prediction points generated
            using either a one-shot- or autoregressive forecast. Setting `n <= output_chunk_length` prevents
            auto-regression. This is useful when the covariates don't extend far enough into the future, or to prohibit
            the model from using future values of past and / or future covariates for prediction (depending on the
            model's covariate support).
        output_chunk_shift
            Optionally, the number of steps to shift the start of the output chunk into the future (relative to the
            input chunk end). This will create a gap between the input and output. If the model supports
            `future_covariates`, the future values are extracted from the shifted output chunk. Predictions will start
            `output_chunk_shift` steps after the end of the target `series`. If `output_chunk_shift` is set, the model
            cannot generate autoregressive predictions (`n > output_chunk_length`).
        d_model
            The number of expected features in the transformer encoder/decoder inputs (default=64).
        nhead
            The number of heads in the multi-head attention mechanism (default=4).
        num_encoder_layers
            The number of encoder layers in the encoder (default=3).
        num_decoder_layers
            The number of decoder layers in the decoder (default=3).
        dim_feedforward
            The dimension of the feedforward network model (default=512).
        dropout
            Fraction of neurons affected by Dropout (default=0.1).
        activation
            The activation function of encoder/decoder intermediate layer, (default='relu').
            can be one of the glu variant's FeedForward Network (FFN)[2]. A feedforward network is a
            fully-connected layer with an activation. The glu variant's FeedForward Network are a series
            of FFNs designed to work better with Transformer based models. ["GLU", "Bilinear", "ReGLU", "GEGLU",
            "SwiGLU", "ReLU", "GELU"] or one the pytorch internal activations ["relu", "gelu"]
        norm_type: str | nn.Module
            The type of LayerNorm variant to use.  Default: ``None``. Available options are
            ["LayerNorm", "RMSNorm", "LayerNormNoBias"], or provide a custom nn.Module.
        custom_encoder
            A custom user-provided encoder module for the transformer (default=None).
        custom_decoder
            A custom user-provided decoder module for the transformer (default=None).
        **kwargs
            Optional arguments to initialize the pytorch_lightning.Module, pytorch_lightning.Trainer, and
            Darts' :class:`TorchForecastingModel`.

        loss_fn
            PyTorch loss function used for training.
            This parameter will be ignored for probabilistic models if the ``likelihood`` parameter is specified.
            Default: ``torch.nn.MSELoss()``.
        likelihood
            One of Darts' :meth:`Likelihood <darts.utils.likelihood_models.torch.TorchLikelihood>` models to be used for
            probabilistic forecasts. Default: ``None``.
        torch_metrics
            A torch metric or a ``MetricCollection`` used for evaluation. A full list of available metrics can be found
            at https://torchmetrics.readthedocs.io/en/latest/. Default: ``None``.
        optimizer_cls
            The PyTorch optimizer class to be used. Default: ``torch.optim.Adam``.
        optimizer_kwargs
            Optionally, some keyword arguments for the PyTorch optimizer (e.g., ``{'lr': 1e-3}``
            for specifying a learning rate). Otherwise the default values of the selected ``optimizer_cls``
            will be used. Default: ``None``.
        lr_scheduler_cls
            Optionally, the PyTorch learning rate scheduler class to be used. Specifying ``None`` corresponds
            to using a constant learning rate. Default: ``None``.
        lr_scheduler_kwargs
            Optionally, some keyword arguments for the PyTorch learning rate scheduler. Default: ``None``.
        use_reversible_instance_norm
            Whether to use reversible instance normalization `RINorm` against distribution shift as shown in [3]_.
            It is only applied to the features of the target series and not the covariates.
        batch_size
            Number of time series (input and output sequences) used in each training pass. Default: ``32``.
        n_epochs
            Number of epochs over which to train the model. Default: ``100``.
        model_name
            Name of the model. Used for creating checkpoints and saving tensorboard data. If not specified,
            defaults to the following string ``"YYYY-mm-dd_HH_MM_SS_torch_model_run_PID"``, where the initial part
            of the name is formatted with the local date and time, while PID is the processed ID (preventing models
            spawned at the same time by different processes to share the same model_name). E.g.,
            ``"2021-06-14_09_53_32_torch_model_run_44607"``.
        work_dir
            Path of the working directory, where to save checkpoints and Tensorboard summaries.
            Default: current working directory.
        log_tensorboard
            If set, use Tensorboard to log the different parameters. The logs will be located in:
            ``"{work_dir}/darts_logs/{model_name}/logs/"``. Default: ``False``.
        nr_epochs_val_period
            Number of epochs to wait before evaluating the validation loss (if a validation
            ``TimeSeries`` is passed to the :func:`fit()` method). Default: ``1``.
        force_reset
            If set to ``True``, any previously-existing model with the same name will be reset (all checkpoints will
            be discarded). Default: ``False``.
        save_checkpoints
            Whether to automatically save the untrained model and checkpoints from training.
            To load the model from checkpoint, call :func:`MyModelClass.load_from_checkpoint()`, where
            :class:`MyModelClass` is the :class:`TorchForecastingModel` class that was used (such as :class:`TFTModel`,
            :class:`NBEATSModel`, etc.). If set to ``False``, the model can still be manually saved using
            :func:`save()` and loaded using :func:`load()`. Default: ``False``.
        add_encoders
            A large number of past and future covariates can be automatically generated with `add_encoders`.
            This can be done by adding multiple pre-defined index encoders and/or custom user-made functions that
            will be used as index encoders. Additionally, a transformer such as Darts' :class:`Scaler` can be added to
            transform the generated covariates. This happens all under one hood and only needs to be specified at
            model creation.
            Read :meth:`SequentialEncoder <darts.dataprocessing.encoders.SequentialEncoder>` to find out more about
            ``add_encoders``. Default: ``None``. An example showing some of ``add_encoders`` features:

            .. highlight:: python
            .. code-block:: python

                def encode_year(idx):
                    return (idx.year - 1950) / 50

                add_encoders={
                    'cyclic': {'future': ['month']},
                    'datetime_attribute': {'future': ['hour', 'dayofweek']},
                    'position': {'past': ['relative'], 'future': ['relative']},
                    'custom': {'past': [encode_year]},
                    'transformer': Scaler(),
                    'tz': 'CET'
                }
            ..
        random_state
            Control the randomness of the weights initialization. Check this
            `link <https://scikit-learn.org/stable/glossary.html#term-random_state>`_ for more details.
            Default: ``None``.
        pl_trainer_kwargs
            By default :class:`TorchForecastingModel` creates a PyTorch Lightning Trainer with several useful presets
            that performs the training, validation and prediction processes. These presets include automatic
            checkpointing, tensorboard logging, setting the torch device and more.
            With ``pl_trainer_kwargs`` you can add additional kwargs to instantiate the PyTorch Lightning trainer
            object. Check the `PL Trainer documentation
            <https://pytorch-lightning.readthedocs.io/en/stable/common/trainer.html>`_ for more information about the
            supported kwargs. Default: ``None``.
            Running on GPU(s) is also possible using ``pl_trainer_kwargs`` by specifying keys ``"accelerator",
            "devices", and "auto_select_gpus"``. Some examples for setting the devices inside the ``pl_trainer_kwargs``
            dict:

            - ``{"accelerator": "cpu"}`` for CPU,
            - ``{"accelerator": "gpu", "devices": [i]}`` to use only GPU ``i`` (``i`` must be an integer),
            - ``{"accelerator": "gpu", "devices": -1, "auto_select_gpus": True}`` to use all available GPUS.

            For more info, see here:
            https://pytorch-lightning.readthedocs.io/en/stable/common/trainer.html#trainer-flags , and
            https://pytorch-lightning.readthedocs.io/en/stable/accelerators/gpu_basic.html#train-on-multiple-gpus

            With parameter ``"callbacks"`` you can add custom or PyTorch-Lightning built-in callbacks to Darts'
            :class:`TorchForecastingModel`. Below is an example for adding EarlyStopping to the training process.
            The model will stop training early if the validation loss `val_loss` does not improve beyond
            specifications. For more information on callbacks, visit:
            `PyTorch Lightning Callbacks
            <https://pytorch-lightning.readthedocs.io/en/stable/extensions/callbacks.html>`_

            .. highlight:: python
            .. code-block:: python

                from pytorch_lightning.callbacks.early_stopping import EarlyStopping

                # stop training when validation loss does not decrease more than 0.05 (`min_delta`) over
                # a period of 5 epochs (`patience`)
                my_stopper = EarlyStopping(
                    monitor="val_loss",
                    patience=5,
                    min_delta=0.05,
                    mode='min',
                )

                pl_trainer_kwargs={"callbacks": [my_stopper]}
            ..

            Note that you can also use a custom PyTorch Lightning Trainer for training and prediction with optional
            parameter ``trainer`` in :func:`fit()` and :func:`predict()`.
        show_warnings
            whether to show warnings raised from PyTorch Lightning. Useful to detect potential issues of
            your forecasting use case. Default: ``False``.

        References
        ----------
        .. [1] Ashish Vaswani, Noam Shazeer, Niki Parmar, Jakob Uszkoreit, Llion Jones, Aidan N Gomez, Lukasz Kaiser,
        and Illia Polosukhin, "Attention Is All You Need", 2017. In Advances in Neural Information Processing Systems,
        pages 6000-6010. https://arxiv.org/abs/1706.03762.
        .. [2] Shazeer, Noam, "GLU Variants Improve Transformer", 2020. arVix https://arxiv.org/abs/2002.05202.
        .. [3] T. Kim et al. "Reversible Instance Normalization for Accurate Time-Series Forecasting against
                Distribution Shift", https://openreview.net/forum?id=cGDAkQo1C0p

        Notes
        -----
        Disclaimer:
        This current implementation is fully functional and can already produce some good predictions. However,
        it is still limited in how it uses the Transformer architecture because the `tgt` input of
        `torch.nn.Transformer` is not utilized to its full extent. Currently, we simply pass the last value of the
        `src` input to `tgt`. To get closer to the way the Transformer is usually used in language models, we
        should allow the model to consume its own output as part of the `tgt` argument, such that when predicting
        sequences of values, the input to the `tgt` argument would grow as outputs of the transformer model would be
        added to it. Of course, the training of the model would have to be adapted accordingly.

        Examples
        --------
        >>> from darts.datasets import WeatherDataset
        >>> from darts.models import TransformerModel
        >>> series = WeatherDataset().load()
        >>> # predicting atmospheric pressure
        >>> target = series['p (mbar)'][:100]
        >>> # optionally, use past observed rainfall (pretending to be unknown beyond index 100)
        >>> past_cov = series['rain (mm)'][:100]
        >>> model = TransformerModel(
        >>>     input_chunk_length=6,
        >>>     output_chunk_length=6,
        >>>     n_epochs=20
        >>> )
        >>> model.fit(target, past_covariates=past_cov)
        >>> pred = model.predict(6)
        >>> pred.values()
        array([[5.40498034],
               [5.36561899],
               [5.80616883],
               [6.48695488],
               [7.63158655],
               [5.65417736]])

        .. note::
            `Transformer example notebook <https://unit8co.github.io/darts/examples/06-Transformer-examples.html>`_
            presents techniques that can be used to improve the forecasts quality compared to this simple usage
            example.
        """
        super().__init__(**self._extract_torch_model_params(**self.model_params))

        # extract pytorch lightning module kwargs
        self.pl_module_params = self._extract_pl_module_params(**self.model_params)

        self.d_model = d_model
        self.nhead = nhead
        self.num_encoder_layers = num_encoder_layers
        self.num_decoder_layers = num_decoder_layers
        self.dim_feedforward = dim_feedforward
        self.dropout = dropout
        self.activation = activation
        self.norm_type = norm_type
        self.custom_encoder = custom_encoder
        self.custom_decoder = custom_decoder

    @property
    def supports_multivariate(self) -> bool:
        return True

    def _create_model(self, train_sample: tuple[torch.Tensor]) -> torch.nn.Module:
        # samples are made of (past_target, past_covariates, future_target)
        input_dim = train_sample[0].shape[1] + (
            train_sample[1].shape[1] if train_sample[1] is not None else 0
        )
        output_dim = train_sample[-1].shape[1]
        nr_params = 1 if self.likelihood is None else self.likelihood.num_parameters

        return _TransformerModule(
            input_size=input_dim,
            output_size=output_dim,
            nr_params=nr_params,
            d_model=self.d_model,
            nhead=self.nhead,
            num_encoder_layers=self.num_encoder_layers,
            num_decoder_layers=self.num_decoder_layers,
            dim_feedforward=self.dim_feedforward,
            dropout=self.dropout,
            activation=self.activation,
            norm_type=self.norm_type,
            custom_encoder=self.custom_encoder,
            custom_decoder=self.custom_decoder,
            **self.pl_module_params,
        )



"""
Likelihoods for `StatsForecast` Models
--------------------------------------
"""

from abc import ABC

import numpy as np

from darts.logging import get_logger
from darts.utils.likelihood_models.base import (
    Likelihood,
    LikelihoodType,
    quantile_names,
)
from darts.utils.utils import _check_quantiles, sample_from_quantiles

logger = get_logger(__name__)


class QuantilePrediction(Likelihood, ABC):
    def __init__(self, quantiles: list[float]):
        """Quantile Prediction Likelihood

        Can be used to generate quantile predictions for any Darts model that wraps a `statsforecast` model.

        Parameters
        ----------
        quantiles
            A list of quantiles. Default: `[0.01, 0.05, 0.1, 0.25, 0.5, 0.75, 0.9, 0.95, 0.99]`.
        """
        if quantiles is None:
            quantiles = [
                0.01,
                0.05,
                0.1,
                0.25,
                0.5,
                0.75,
                0.9,
                0.95,
                0.99,
            ]
        else:
            quantiles = sorted(quantiles)
            _check_quantiles(quantiles)
        self.quantiles = quantiles
        self._median_idx = quantiles.index(0.5)
        self.levels = [
            round(100 * (q_h - q_l), 2)
            for q_l, q_h in zip(
                quantiles[: self._median_idx], quantiles[self._median_idx + 1 :][::-1]
            )
        ]

        super().__init__(
            likelihood_type=LikelihoodType.Quantile,
            parameter_names=quantile_names(self.quantiles),
        )
        # ignore additional attrs for equality tests
        self.ignore_attrs_equality += ["_median_idx", "levels"]

    def predict(self, model_output: np.ndarray, num_samples: int) -> np.ndarray:
        """
        Generates sampled or direct likelihood parameter predictions.

        Parameters
        ----------
        model_output
            The output of the StatsForecast model.
        num_samples
            Number of times a prediction is sampled from the likelihood model / distribution.
            If `1` and `predict_likelihood_parameters=False`, returns median / mean predictions.
        """
        if num_samples > 1:
            return self.sample(model_output, num_samples=num_samples)

        # likelihood parameters or median prediction are handled by the `levels` parameter in
        # `StatsForecastModel._predict()`
        return model_output

    def sample(self, model_output: np.ndarray, num_samples: int) -> np.ndarray:
        """
        Samples a prediction from the likelihood distribution and the predicted parameters.
        """
        return sample_from_quantiles(
            vals=model_output,
            quantiles=np.array(self.quantiles),
            num_samples=num_samples,
        )



"""
CatBoost model
--------------

CatBoost based regression model.

This implementation comes with the ability to produce probabilistic forecasts.
"""

from collections.abc import Sequence
from typing import Any, Optional, Union

import numpy as np
import pandas as pd
from catboost import CatBoostRegressor, Pool

from darts.logging import get_logger
from darts.models.forecasting.regression_model import (
    RegressionModelWithCategoricalCovariates,
    _QuantileModelContainer,
)
from darts.timeseries import TimeSeries
from darts.utils.likelihood_models.sklearn import (
    QuantileRegression,
    _check_likelihood,
    _get_likelihood,
)

logger = get_logger(__name__)


class CatBoostModel(RegressionModelWithCategoricalCovariates):
    def __init__(
        self,
        lags: Union[int, list] = None,
        lags_past_covariates: Union[int, list[int]] = None,
        lags_future_covariates: Union[tuple[int, int], list[int]] = None,
        output_chunk_length: int = 1,
        output_chunk_shift: int = 0,
        add_encoders: Optional[dict] = None,
        likelihood: Optional[str] = None,
        quantiles: list = None,
        random_state: Optional[int] = None,
        multi_models: Optional[bool] = True,
        use_static_covariates: bool = True,
        categorical_past_covariates: Optional[Union[str, list[str]]] = None,
        categorical_future_covariates: Optional[Union[str, list[str]]] = None,
        categorical_static_covariates: Optional[Union[str, list[str]]] = None,
        **kwargs,
    ):
        """CatBoost Model

        Parameters
        ----------
        lags
            Lagged target `series` values used to predict the next time step/s.
            If an integer, must be > 0. Uses the last `n=lags` past lags; e.g. `(-1, -2, ..., -lags)`, where `0`
            corresponds the first predicted time step of each sample. If `output_chunk_shift > 0`, then
            lag `-1` translates to `-1 - output_chunk_shift` steps before the first prediction step.
            If a list of integers, each value must be < 0. Uses only the specified values as lags.
            If a dictionary, the keys correspond to the `series` component names (of the first series when
            using multiple series) and the values correspond to the component lags (integer or list of integers). The
            key 'default_lags' can be used to provide default lags for un-specified components. Raises and error if some
            components are missing and the 'default_lags' key is not provided.
        lags_past_covariates
            Lagged `past_covariates` values used to predict the next time step/s.
            If an integer, must be > 0. Uses the last `n=lags_past_covariates` past lags; e.g. `(-1, -2, ..., -lags)`,
            where `0` corresponds to the first predicted time step of each sample. If `output_chunk_shift > 0`, then
            lag `-1` translates to `-1 - output_chunk_shift` steps before the first prediction step.
            If a list of integers, each value must be < 0. Uses only the specified values as lags.
            If a dictionary, the keys correspond to the `past_covariates` component names (of the first series when
            using multiple series) and the values correspond to the component lags (integer or list of integers). The
            key 'default_lags' can be used to provide default lags for un-specified components. Raises and error if some
            components are missing and the 'default_lags' key is not provided.
        lags_future_covariates
            Lagged `future_covariates` values used to predict the next time step/s. The lags are always relative to the
            first step in the output chunk, even when `output_chunk_shift > 0`.
            If a tuple of `(past, future)`, both values must be > 0. Uses the last `n=past` past lags and `n=future`
            future lags; e.g. `(-past, -(past - 1), ..., -1, 0, 1, .... future - 1)`, where `0` corresponds the first
            predicted time step of each sample. If `output_chunk_shift > 0`, the position of negative lags differ from
            those of `lags` and `lags_past_covariates`. In this case a future lag `-5` would point at the same
            step as a target lag of `-5 + output_chunk_shift`.
            If a list of integers, uses only the specified values as lags.
            If a dictionary, the keys correspond to the `future_covariates` component names (of the first series when
            using multiple series) and the values correspond to the component lags (tuple or list of integers). The key
            'default_lags' can be used to provide default lags for un-specified components. Raises and error if some
            components are missing and the 'default_lags' key is not provided.
        output_chunk_length
            Number of time steps predicted at once (per chunk) by the internal model. It is not the same as forecast
            horizon `n` used in `predict()`, which is the desired number of prediction points generated using a
            one-shot- or autoregressive forecast. Setting `n <= output_chunk_length` prevents auto-regression. This is
            useful when the covariates don't extend far enough into the future, or to prohibit the model from using
            future values of past and / or future covariates for prediction (depending on the model's covariate
            support).
        output_chunk_shift
            Optionally, the number of steps to shift the start of the output chunk into the future (relative to the
            input chunk end). This will create a gap between the input (history of target and past covariates) and
            output. If the model supports `future_covariates`, the `lags_future_covariates` are relative to the first
            step in the shifted output chunk. Predictions will start `output_chunk_shift` steps after the end of the
            target `series`. If `output_chunk_shift` is set, the model cannot generate autoregressive predictions
            (`n > output_chunk_length`).
        add_encoders
            A large number of past and future covariates can be automatically generated with `add_encoders`.
            This can be done by adding multiple pre-defined index encoders and/or custom user-made functions that
            will be used as index encoders. Additionally, a transformer such as Darts' :class:`Scaler` can be added to
            transform the generated covariates. This happens all under one hood and only needs to be specified at
            model creation.
            Read :meth:`SequentialEncoder <darts.dataprocessing.encoders.SequentialEncoder>` to find out more about
            ``add_encoders``. Default: ``None``. An example showing some of ``add_encoders`` features:

            .. highlight:: python
            .. code-block:: python

                def encode_year(idx):
                    return (idx.year - 1950) / 50

                add_encoders={
                    'cyclic': {'future': ['month']},
                    'datetime_attribute': {'future': ['hour', 'dayofweek']},
                    'position': {'past': ['relative'], 'future': ['relative']},
                    'custom': {'past': [encode_year]},
                    'transformer': Scaler(),
                    'tz': 'CET'
                }
            ..
        likelihood
            Can be set to 'quantile', 'poisson' or 'gaussian'. If set, the model will be probabilistic,
            allowing sampling at prediction time. When set to 'gaussian', the model will use CatBoost's
            'RMSEWithUncertainty' loss function. When using this loss function, CatBoost returns a mean
            and variance couple, which capture data (aleatoric) uncertainty.
            This will overwrite any `objective` parameter.
        quantiles
            Fit the model to these quantiles if the `likelihood` is set to `quantile`.
        random_state
            Control the randomness in the fitting procedure and for sampling.
            Default: ``None``.
        multi_models
            If True, a separate model will be trained for each future lag to predict. If False, a single model
            is trained to predict all the steps in 'output_chunk_length' (features lags are shifted back by
            `output_chunk_length - n` for each step `n`). Default: True.
        use_static_covariates
            Whether the model should use static covariate information in case the input `series` passed to ``fit()``
            contain static covariates. If ``True``, and static covariates are available at fitting time, will enforce
            that all target `series` have the same static covariate dimensionality in ``fit()`` and ``predict()``.
        categorical_past_covariates
            Optionally, component name or list of component names specifying the past covariates that should be treated
            as categorical by the underlying `CatBoostRegressor`. The components that are specified as categorical
            must be integer-encoded. For more information on how CatBoost handles categorical features,
            visit: `Categorical feature support documentatio
            <https://catboost.ai/docs/en/features/categorical-features>`_.
        categorical_future_covariates
            Optionally, component name or list of component names specifying the future covariates that should be
            treated as categorical by the underlying `CatBoostRegressor`. The components that
            are specified as categorical must be integer-encoded.
        categorical_static_covariates
            Optionally, string or list of strings specifying the static covariates that should be treated as categorical
            by the underlying `CatBoostRegressor`. The components that
            are specified as categorical must be integer-encoded.
        **kwargs
            Additional keyword arguments passed to `catboost.CatBoostRegressor`.
            Native multi-output support can be achieved by using an appropriate `loss_function` ('MultiRMSE',
            'MultiRMSEWithMissingValues'). Otherwise, Darts uses its `MultiOutputRegressor` wrapper to add multi-output
            support.

        Examples
        --------
        >>> from darts.datasets import WeatherDataset
        >>> from darts.models import CatBoostModel
        >>> series = WeatherDataset().load()
        >>> # predicting atmospheric pressure
        >>> target = series['p (mbar)'][:100]
        >>> # optionally, use past observed rainfall (pretending to be unknown beyond index 100)
        >>> past_cov = series['rain (mm)'][:100]
        >>> # optionally, use future temperatures (pretending this component is a forecast)
        >>> future_cov = series['T (degC)'][:106]
        >>> # predict 6 pressure values using the 12 past values of pressure and rainfall, as well as the 6 temperature
        >>> # values corresponding to the forecasted period
        >>> model = CatBoostModel(
        >>>     lags=12,
        >>>     lags_past_covariates=12,
        >>>     lags_future_covariates=[0,1,2,3,4,5],
        >>>     output_chunk_length=6
        >>> )
        >>> model.fit(target, past_covariates=past_cov, future_covariates=future_cov)
        >>> pred = model.predict(6)
        >>> pred.values()
        array([[1006.4153701 ],
               [1006.41907237],
               [1006.30872957],
               [1006.28614154],
               [1006.22355514],
               [1006.21607546]])
        """
        kwargs["random_state"] = random_state  # seed for tree learner
        self.kwargs = kwargs
        self._model_container = None

        # parse likelihood
        if likelihood is not None:
            likelihood_map = {
                "quantile": None,
                "poisson": "Poisson",
                "gaussian": "RMSEWithUncertainty",
                "RMSEWithUncertainty": "RMSEWithUncertainty",
            }
            _check_likelihood(likelihood, list(likelihood_map.keys()))
            if likelihood == "RMSEWithUncertainty":
                # RMSEWithUncertainty returns mean and variance which is equivalent to gaussian
                likelihood = "gaussian"

            if likelihood == "quantile":
                self._model_container = _QuantileModelContainer()
            else:
                self.kwargs["loss_function"] = likelihood_map[likelihood]

        self._likelihood = _get_likelihood(
            likelihood=likelihood,
            n_outputs=output_chunk_length if multi_models else 1,
            random_state=random_state,
            quantiles=quantiles,
        )

        # suppress writing catboost info files when user does not specifically ask to
        if "allow_writing_files" not in kwargs:
            kwargs["allow_writing_files"] = False

        super().__init__(
            lags=lags,
            lags_past_covariates=lags_past_covariates,
            lags_future_covariates=lags_future_covariates,
            output_chunk_length=output_chunk_length,
            output_chunk_shift=output_chunk_shift,
            add_encoders=add_encoders,
            multi_models=multi_models,
            model=CatBoostRegressor(**kwargs),
            use_static_covariates=use_static_covariates,
            categorical_past_covariates=categorical_past_covariates,
            categorical_future_covariates=categorical_future_covariates,
            categorical_static_covariates=categorical_static_covariates,
        )

        # if no loss provided, get the default loss from the model
        self.kwargs["loss_function"] = self.model.get_params().get("loss_function")

    def fit(
        self,
        series: Union[TimeSeries, Sequence[TimeSeries]],
        past_covariates: Optional[Union[TimeSeries, Sequence[TimeSeries]]] = None,
        future_covariates: Optional[Union[TimeSeries, Sequence[TimeSeries]]] = None,
        val_series: Optional[Union[TimeSeries, Sequence[TimeSeries]]] = None,
        val_past_covariates: Optional[Union[TimeSeries, Sequence[TimeSeries]]] = None,
        val_future_covariates: Optional[Union[TimeSeries, Sequence[TimeSeries]]] = None,
        max_samples_per_ts: Optional[int] = None,
        n_jobs_multioutput_wrapper: Optional[int] = None,
        sample_weight: Optional[Union[TimeSeries, Sequence[TimeSeries], str]] = None,
        val_sample_weight: Optional[
            Union[TimeSeries, Sequence[TimeSeries], str]
        ] = None,
        verbose: Optional[Union[int, bool]] = 0,
        **kwargs,
    ):
        """
        Fits/trains the model using the provided list of features time series and the target time series.

        Parameters
        ----------
        series
            TimeSeries or Sequence[TimeSeries] object containing the target values.
        past_covariates
            Optionally, a series or sequence of series specifying past-observed covariates
        future_covariates
            Optionally, a series or sequence of series specifying future-known covariates
        val_series
            TimeSeries or Sequence[TimeSeries] object containing the target values for evaluation dataset
        val_past_covariates
            Optionally, a series or sequence of series specifying past-observed covariates for evaluation dataset
        val_future_covariates : Union[TimeSeries, Sequence[TimeSeries]]
            Optionally, a series or sequence of series specifying future-known covariates for evaluation dataset
        max_samples_per_ts
            This is an integer upper bound on the number of tuples that can be produced
            per time series. It can be used in order to have an upper bound on the total size of the dataset and
            ensure proper sampling. If `None`, it will read all of the individual time series in advance (at dataset
            creation) to know their sizes, which might be expensive on big datasets.
            If some series turn out to have a length that would allow more than `max_samples_per_ts`, only the
            most recent `max_samples_per_ts` samples will be considered.
        n_jobs_multioutput_wrapper
            Number of jobs of the MultiOutputRegressor wrapper to run in parallel. Only used if the model doesn't
            support multi-output regression natively.
        sample_weight
            Optionally, some sample weights to apply to the target `series` labels. They are applied per observation,
            per label (each step in `output_chunk_length`), and per component.
            If a series or sequence of series, then those weights are used. If the weight series only have a single
            component / column, then the weights are applied globally to all components in `series`. Otherwise, for
            component-specific weights, the number of components must match those of `series`.
            If a string, then the weights are generated using built-in weighting functions. The available options are
            `"linear"` or `"exponential"` decay - the further in the past, the lower the weight. The weights are
            computed globally based on the length of the longest series in `series`. Then for each series, the weights
            are extracted from the end of the global weights. This gives a common time weighting across all series.
        val_sample_weight
            Same as for `sample_weight` but for the evaluation dataset.
        verbose
            An integer or a boolean that can be set to 1 to display catboost's default verbose output
        **kwargs
            Additional kwargs passed to `catboost.CatboostRegressor.fit()`
        """
        likelihood = self.likelihood
        if isinstance(likelihood, QuantileRegression):
            # empty model container in case of multiple calls to fit, e.g. when backtesting
            self._model_container.clear()
            for quantile in likelihood.quantiles:
                this_quantile = str(quantile)
                # translating to catboost argument
                self.kwargs["loss_function"] = f"Quantile:alpha={this_quantile}"
                self.model = CatBoostRegressor(**self.kwargs)
                super().fit(
                    series=series,
                    past_covariates=past_covariates,
                    future_covariates=future_covariates,
                    val_series=val_series,
                    val_past_covariates=val_past_covariates,
                    val_future_covariates=val_future_covariates,
                    max_samples_per_ts=max_samples_per_ts,
                    n_jobs_multioutput_wrapper=n_jobs_multioutput_wrapper,
                    sample_weight=sample_weight,
                    val_sample_weight=val_sample_weight,
                    verbose=verbose,
                    **kwargs,
                )
                self._model_container[quantile] = self.model
            return self

        super().fit(
            series=series,
            past_covariates=past_covariates,
            future_covariates=future_covariates,
            val_series=val_series,
            val_past_covariates=val_past_covariates,
            val_future_covariates=val_future_covariates,
            max_samples_per_ts=max_samples_per_ts,
            n_jobs_multioutput_wrapper=n_jobs_multioutput_wrapper,
            sample_weight=sample_weight,
            val_sample_weight=val_sample_weight,
            verbose=verbose,
            **kwargs,
        )
        return self

    def _add_val_set_to_kwargs(
        self,
        kwargs: dict,
        val_series: Sequence[TimeSeries],
        val_past_covariates: Optional[Sequence[TimeSeries]],
        val_future_covariates: Optional[Sequence[TimeSeries]],
        val_sample_weight: Optional[Union[Sequence[TimeSeries], str]],
        max_samples_per_ts: int,
    ) -> dict:
        # CatBoostRegressor requires sample weights to be passed with a validation set `Pool`
        kwargs = super()._add_val_set_to_kwargs(
            kwargs=kwargs,
            val_series=val_series,
            val_past_covariates=val_past_covariates,
            val_future_covariates=val_future_covariates,
            val_sample_weight=val_sample_weight,
            max_samples_per_ts=max_samples_per_ts,
        )
        val_set_name, val_weight_name = self.val_set_params
        val_sets = kwargs[val_set_name]
        # CatBoost requires eval set Pool with sample weights -> remove from kwargs
        val_weights = kwargs.pop(val_weight_name)
        val_pools = []
        for i, val_set in enumerate(val_sets):
            val_pools.append(
                Pool(
                    data=val_set[0],
                    label=val_set[1],
                    weight=val_weights[i] if val_weights is not None else None,
                    cat_features=self._categorical_indices,
                )
            )
        kwargs[val_set_name] = val_pools
        return kwargs

    @property
    def supports_val_set(self) -> bool:
        return True

    @property
    def val_set_params(self) -> tuple[Optional[str], Optional[str]]:
        return "eval_set", "eval_sample_weight"

    @property
    def min_train_series_length(self) -> int:
        # Catboost requires a minimum of 2 train samples, therefore the min_train_series_length should be one more than
        # for other regression models
        return max(
            3,
            (
                -self.lags["target"][0] + self.output_chunk_length + 1
                if "target" in self.lags
                else self.output_chunk_length
            ),
        )

    @property
    def _supports_native_multioutput(self):
        # CatBoostRegressor supports multi-output natively, but only with selected loss functions
        # ("MultiRMSE", "MultiRMSEWithMissingValues", ...)
        return CatBoostRegressor._is_multiregression_objective(
            self.kwargs.get("loss_function")
        )

    @property
    def _categorical_fit_param(self) -> Optional[str]:
        """
        Returns the name of the categorical features parameter from model's `fit` method .
        """
        return "cat_features"

    def _format_samples(
        self, samples: np.ndarray, labels: Optional[np.ndarray] = None
    ) -> tuple[Any, Any]:
        """
        CatBoost currently only supports categorical features as int.
        If categorical features are specified, the samples are converted into a pandas DataFrame and categorical
        columns are cast to integer.
        """
        samples, labels = super()._format_samples(samples, labels=labels)
        if len(self._categorical_indices) != 0:
            # transform into pandas df and cast categorical columns to int
            samples = pd.DataFrame(samples)
            samples = samples.astype({col: int for col in self._categorical_indices})
        return samples, labels



import numpy as np
import pytest

from darts.tests.conftest import TORCH_AVAILABLE, tfm_kwargs
from darts.utils.timeseries_generation import linear_timeseries

if not TORCH_AVAILABLE:
    pytest.skip(
        f"Torch not available. {__name__} tests will be skipped.",
        allow_module_level=True,
    )
import pytorch_lightning as pl

from darts.models.forecasting.rnn_model import RNNModel


class TestPTLTrainer:
    trainer_params = {
        "max_epochs": 1,
        "logger": False,
        "enable_checkpointing": False,
    }
    series = linear_timeseries(length=100).astype(np.float32)
    pl_200_or_above = int(pl.__version__.split(".")[0]) >= 2
    precisions = {
        32: "32" if not pl_200_or_above else "32-true",
        64: "64" if not pl_200_or_above else "64-true",
    }

    def test_prediction_loaded_custom_trainer(self, tmpdir_module):
        """validate manual save with automatic save files by comparing output between the two"""
        auto_name = "test_save_automatic"
        model = RNNModel(
            12,
            "RNN",
            10,
            10,
            model_name=auto_name,
            work_dir=tmpdir_module,
            save_checkpoints=True,
            random_state=42,
            **tfm_kwargs,
        )

        # fit model with custom trainer
        trainer = pl.Trainer(
            max_epochs=1,
            enable_checkpointing=True,
            logger=False,
            callbacks=model.trainer_params["callbacks"],
            **tfm_kwargs["pl_trainer_kwargs"],
        )
        model.fit(self.series, trainer=trainer)

        # load automatically saved model with manual load_model() and load_from_checkpoint()
        model_loaded = RNNModel.load_from_checkpoint(
            model_name=auto_name,
            work_dir=tmpdir_module,
            best=False,
            map_location="cpu",
        )

        # compare prediction of loaded model with original model
        assert model.predict(n=4) == model_loaded.predict(n=4)

    def test_prediction_custom_trainer(self):
        model = RNNModel(12, "RNN", 10, 10, random_state=42, **tfm_kwargs)
        model2 = RNNModel(12, "RNN", 10, 10, random_state=42, **tfm_kwargs)

        # fit model with custom trainer
        trainer = pl.Trainer(
            **self.trainer_params,
            precision=self.precisions[32],
            **tfm_kwargs["pl_trainer_kwargs"],
        )
        model.fit(self.series, trainer=trainer)

        # fit model with built-in trainer
        model2.fit(self.series, epochs=1)

        # both should produce identical prediction
        assert model.predict(n=4) == model2.predict(n=4)

    def test_custom_trainer_setup(self):
        model = RNNModel(12, "RNN", 10, 10, random_state=42, **tfm_kwargs)

        # trainer with wrong precision should raise ValueError
        trainer = pl.Trainer(
            **self.trainer_params,
            precision=self.precisions[64],
            **tfm_kwargs["pl_trainer_kwargs"],
        )
        with pytest.raises(ValueError):
            model.fit(self.series, trainer=trainer)

        # no error with correct precision
        trainer = pl.Trainer(
            **self.trainer_params,
            precision=self.precisions[32],
            **tfm_kwargs["pl_trainer_kwargs"],
        )
        model.fit(self.series, trainer=trainer)

        # check if number of epochs trained is same as trainer.max_epochs
        assert trainer.max_epochs == model.epochs_trained

    def test_builtin_extended_trainer(self):
        # wrong precision parameter name
        with pytest.raises(TypeError):
            invalid_trainer_kwarg = {
                "precisionn": self.precisions[32],
                **tfm_kwargs["pl_trainer_kwargs"],
            }
            model = RNNModel(
                12,
                "RNN",
                10,
                10,
                random_state=42,
                pl_trainer_kwargs=invalid_trainer_kwarg,
            )
            model.fit(self.series, epochs=1)

        # float 16 not supported
        with pytest.raises(ValueError):
            invalid_trainer_kwarg = {
                "precision": "16-mixed",
                **tfm_kwargs["pl_trainer_kwargs"],
            }
            model = RNNModel(
                12,
                "RNN",
                10,
                10,
                random_state=42,
                pl_trainer_kwargs=invalid_trainer_kwarg,
            )
            model.fit(self.series.astype(np.float16), epochs=1)

        # precision value doesn't match `series` dtype
        with pytest.raises(ValueError):
            invalid_trainer_kwarg = {
                "precision": self.precisions[64],
                **tfm_kwargs["pl_trainer_kwargs"],
            }
            model = RNNModel(
                12,
                "RNN",
                10,
                10,
                random_state=42,
                pl_trainer_kwargs=invalid_trainer_kwarg,
            )
            model.fit(self.series.astype(np.float32), epochs=1)

        for precision in [64, 32]:
            valid_trainer_kwargs = {
                "precision": self.precisions[precision],
                **tfm_kwargs["pl_trainer_kwargs"],
            }

            # valid parameters shouldn't raise error
            model = RNNModel(
                12,
                "RNN",
                10,
                10,
                random_state=42,
                pl_trainer_kwargs=valid_trainer_kwargs,
            )
            ts_dtype = getattr(np, f"float{precision}")
            model.fit(self.series.astype(ts_dtype), epochs=1)
            preds = model.predict(n=3)
            assert model.trainer.precision == self.precisions[precision]
            assert preds.dtype == ts_dtype

    def test_custom_callback(self, tmpdir_module):
        class CounterCallback(pl.callbacks.Callback):
            # counts the number of trained epochs starting from count_default
            def __init__(self, count_default):
                self.counter = count_default

            def on_train_epoch_end(self, *args, **kwargs):
                self.counter += 1

        my_counter_0 = CounterCallback(count_default=0)
        my_counter_2 = CounterCallback(count_default=2)

        model = RNNModel(
            12,
            "RNN",
            10,
            10,
            random_state=42,
            pl_trainer_kwargs={
                "callbacks": [my_counter_0, my_counter_2],
                **tfm_kwargs["pl_trainer_kwargs"],
            },
        )

        # check if callbacks were added
        assert len(model.trainer_params["callbacks"]) == 2
        model.fit(self.series, epochs=2, verbose=True)
        # check that lightning did not mutate callbacks (verbosity adds a progress bar callback)
        assert len(model.trainer_params["callbacks"]) == 2

        assert my_counter_0.counter == model.epochs_trained
        assert my_counter_2.counter == model.epochs_trained + 2

        # check that callbacks don't overwrite Darts' built-in checkpointer
        model = RNNModel(
            12,
            "RNN",
            10,
            10,
            random_state=42,
            work_dir=tmpdir_module,
            save_checkpoints=True,
            pl_trainer_kwargs={
                "callbacks": [CounterCallback(0), CounterCallback(2)],
                **tfm_kwargs["pl_trainer_kwargs"],
            },
        )
        # we expect 3 callbacks
        assert len(model.trainer_params["callbacks"]) == 3

        # first one is our Checkpointer
        assert isinstance(
            model.trainer_params["callbacks"][0], pl.callbacks.ModelCheckpoint
        )

        # second and third are CounterCallbacks
        for i in range(1, 3):
            assert isinstance(model.trainer_params["callbacks"][i], CounterCallback)

    def test_early_stopping(self):
        my_stopper = pl.callbacks.early_stopping.EarlyStopping(
            monitor="val_loss",
            stopping_threshold=1e9,
        )
        model = RNNModel(
            12,
            "RNN",
            10,
            10,
            nr_epochs_val_period=1,
            random_state=42,
            pl_trainer_kwargs={
                "callbacks": [my_stopper],
                **tfm_kwargs["pl_trainer_kwargs"],
            },
        )

        # training should stop immediately with high stopping_threshold
        model.fit(self.series, val_series=self.series, epochs=100, verbose=True)
        assert model.epochs_trained == 1

        # check that early stopping only takes valid monitor variables
        my_stopper = pl.callbacks.early_stopping.EarlyStopping(
            monitor="invalid_variable",
            stopping_threshold=1e9,
        )
        model = RNNModel(
            12,
            "RNN",
            10,
            10,
            nr_epochs_val_period=1,
            random_state=42,
            pl_trainer_kwargs={
                "callbacks": [my_stopper],
                **tfm_kwargs["pl_trainer_kwargs"],
            },
        )

        with pytest.raises(RuntimeError):
            model.fit(self.series, val_series=self.series, epochs=100, verbose=True)



"""
Likelihoods for `RegressionModel`
---------------------------------
"""

from abc import ABC, abstractmethod
from typing import Optional

import numpy as np

from darts.logging import get_logger, raise_log
from darts.utils.likelihood_models.base import (
    Likelihood,
    LikelihoodType,
    quantile_names,
)
from darts.utils.utils import _check_quantiles

logger = get_logger(__name__)


class SKLearnLikelihood(Likelihood, ABC):
    def __init__(
        self,
        likelihood_type: LikelihoodType,
        parameter_names: list[str],
        n_outputs: int,
        random_state: Optional[int] = None,
    ):
        """Base class for sklearn wrapper (e.g. `RegressionModel`) likelihoods.

        Parameters
        ----------
        likelihood_type
            A pre-defined `LikelihoodType`.
        parameter_names
            The likelihood (distribution) parameter names.
        n_outputs
            The number of predicted outputs per model call. `1` if `multi_models=False`, otherwise
            `output_chunk_length`.
        random_state
            Optionally, control the randomness of the sampling.
        """
        self._n_outputs = n_outputs
        self._rng = np.random.default_rng(seed=random_state)
        super().__init__(
            likelihood_type=likelihood_type,
            parameter_names=parameter_names,
        )
        # ignore additional attrs for equality tests
        self.ignore_attrs_equality += ["_n_outputs", "_rng"]

    def predict(
        self,
        model,
        x: np.ndarray,
        num_samples: int,
        predict_likelihood_parameters: bool,
        **kwargs,
    ) -> np.ndarray:
        """
        Generates sampled or direct likelihood parameter predictions.

        Parameters
        ----------
        model
            The Darts `RegressionModel`.
        x
            The input feature array passed to the underlying estimator's `predict()` method.
        num_samples
            Number of times a prediction is sampled from the likelihood model / distribution.
            If `1` and `predict_likelihood_parameters=False`, returns median / mean predictions.
        predict_likelihood_parameters
            If set to `True`, generates likelihood parameter predictions instead of sampling from the
            likelihood model / distribution. Only supported with `num_samples = 1` and
            `n<=output_chunk_length`.
        kwargs
            Some kwargs passed to the underlying estimator's `predict()` method.
        """
        model_output = self._estimator_predict(model, x=x, **kwargs)
        if predict_likelihood_parameters:
            return self.predict_likelihood_parameters(model_output)
        elif num_samples == 1:
            return self._get_median_prediction(model_output)
        else:
            return self.sample(model_output)

    @abstractmethod
    def sample(self, model_output: np.ndarray) -> np.ndarray:
        """
        Samples a prediction from the likelihood distribution and the predicted parameters.
        """

    @abstractmethod
    def predict_likelihood_parameters(self, model_output: np.ndarray) -> np.ndarray:
        """
        Returns the distribution parameters as a array, extracted from the raw model outputs.
        """

    @abstractmethod
    def _estimator_predict(
        self,
        model,
        x: np.ndarray,
        **kwargs,
    ) -> np.ndarray:
        """
        Computes the model output.

        Parameters
        ----------
        model
            The Darts `RegressionModel`.
        x
            The input feature array passed to the underlying estimator's `predict()` method.
        kwargs
            Some kwargs passed to the underlying estimator's `predict()` method.
        """

    @abstractmethod
    def _get_median_prediction(self, model_output: np.ndarray) -> np.ndarray:
        """
        Gets the median prediction per component extracted from the model output.
        """


class GaussianLikelihood(SKLearnLikelihood):
    def __init__(
        self,
        n_outputs: int,
        random_state: Optional[int] = None,
    ):
        """
        Gaussian distribution [1]_.

        Parameters
        ----------
        n_outputs
            The number of predicted outputs per model call. `1` if `multi_models=False`, otherwise
            `output_chunk_length`.
        random_state
            Optionally, control the randomness of the sampling.

        References
        ----------
        .. [1] https://en.wikipedia.org/wiki/Normal_distribution
        """
        super().__init__(
            likelihood_type=LikelihoodType.Gaussian,
            parameter_names=["mu", "sigma"],
            n_outputs=n_outputs,
            random_state=random_state,
        )

    def sample(self, model_output: np.ndarray) -> np.ndarray:
        # shape (n_components * output_chunk_length, n_series * n_samples, 2)
        # [mu, sigma] on the last dimension, grouped by component
        n_entries, n_samples, n_params = model_output.shape

        # get samples (n_components * output_chunk_length, n_series * n_samples)
        samples = self._rng.normal(
            model_output[:, :, 0],  # mean
            model_output[:, :, 1],  # variance
        )
        # reshape to (n_series * n_samples, n_components * output_chunk_length)
        samples = samples.transpose()
        # reshape to (n_series * n_samples, output_chunk_length, n_components)
        samples = samples.reshape(n_samples, self._n_outputs, -1)
        return samples

    def predict_likelihood_parameters(self, model_output: np.ndarray) -> np.ndarray:
        # shape (n_components * output_chunk_length, n_series * n_samples, 2)
        # [mu, sigma] on the last dimension, grouped by component
        n_samples = model_output.shape[1]

        # reshape to (n_series * n_samples, output_chunk_length, n_components)
        params_reshaped = model_output.transpose(1, 0, 2).reshape(
            n_samples, self._n_outputs, -1
        )
        return params_reshaped

    def _estimator_predict(
        self,
        model,
        x: np.ndarray,
        **kwargs,
    ) -> np.ndarray:
        # returns samples computed from double-valued inputs [mean, variance].
        # `x` is of shape (n_series * n_samples, n_regression_features)
        # `model_output` is of shape:
        #  - (n_series * n_samples, 2): if univariate & output_chunk_length == 1
        #  - (2, n_series * n_samples, n_components * output_chunk_length): otherwise
        # where the axis with 2 dims is mu, sigma
        model_output = model.model.predict(x, **kwargs)
        output_dim = len(model_output.shape)

        # univariate & single-chunk output
        if output_dim <= 2:
            # embedding well shaped 2D output into 3D
            model_output = np.expand_dims(model_output, axis=0)
        else:
            # we transpose to get mu, sigma couples on last axis
            # shape becomes: (n_components * output_chunk_length, n_series * n_samples, 2)
            model_output = model_output.transpose()

        # shape (n_components * output_chunk_length, n_series * n_samples, 2)
        return model_output

    def _get_median_prediction(self, model_output: np.ndarray) -> np.ndarray:
        # shape (n_components * output_chunk_length, n_series * n_samples, 2)
        # [mu, sigma] on the last dimension, grouped by component
        k = model_output.shape[1]
        # extract mu (mean) per component
        component_medians = slice(0, None, self.num_parameters)
        # shape (n_series * n_samples, output_chunk_length, n_components)
        return model_output[:, :, component_medians].reshape(k, self._n_outputs, -1)


class PoissonLikelihood(SKLearnLikelihood):
    def __init__(
        self,
        n_outputs: int,
        random_state: Optional[int] = None,
    ):
        """
        Poisson distribution [1]_.

        Parameters
        ----------
        n_outputs
            The number of predicted outputs per model call. `1` if `multi_models=False`, otherwise
            `output_chunk_length`.
        random_state
            Optionally, control the randomness of the sampling.

        References
        ----------
        .. [1] https://en.wikipedia.org/wiki/Poisson_distribution
        """
        super().__init__(
            likelihood_type=LikelihoodType.Poisson,
            parameter_names=["lambda"],
            n_outputs=n_outputs,
            random_state=random_state,
        )

    def sample(self, model_output: np.ndarray) -> np.ndarray:
        # shape (n_series * n_samples, output_chunk_length, n_components)
        return self._rng.poisson(lam=model_output).astype(float)

    def predict_likelihood_parameters(self, model_output: np.ndarray) -> np.ndarray:
        # lambdas on the last dimension, grouped by component
        return model_output

    def _estimator_predict(
        self,
        model,
        x: np.ndarray,
        **kwargs,
    ) -> np.ndarray:
        k = x.shape[0]
        # returns shape (n_series * n_samples, output_chunk_length, n_components)
        return model.model.predict(x, **kwargs).reshape(k, self._n_outputs, -1)

    def _get_median_prediction(self, model_output: np.ndarray) -> np.ndarray:
        # shape (n_series * n_samples, output_chunk_length, n_components)
        # lambda is already the median prediction
        return model_output


class QuantileRegression(SKLearnLikelihood):
    def __init__(
        self,
        n_outputs: int,
        random_state: Optional[int] = None,
        quantiles: Optional[list[float]] = None,
    ):
        """
        Quantile Regression [1]_.

        Parameters
        ----------
        n_outputs
            The number of predicted outputs per model call. `1` if `multi_models=False`, otherwise
            `output_chunk_length`.
        random_state
            Optionally, control the randomness of the sampling.
        quantiles
            A list of quantiles. Default: `[0.01, 0.05, 0.1, 0.25, 0.5, 0.75, 0.9, 0.95, 0.99]`.

        References
        ----------
        .. [1] https://en.wikipedia.org/wiki/Quantile_regression
        """
        if quantiles is None:
            quantiles = [
                0.01,
                0.05,
                0.1,
                0.25,
                0.5,
                0.75,
                0.9,
                0.95,
                0.99,
            ]
        else:
            quantiles = sorted(quantiles)
            _check_quantiles(quantiles)
        self.quantiles = quantiles
        self._median_idx = quantiles.index(0.5)

        super().__init__(
            likelihood_type=LikelihoodType.Quantile,
            parameter_names=quantile_names(self.quantiles),
            n_outputs=n_outputs,
            random_state=random_state,
        )
        self.ignore_attrs_equality += ["_median_idx"]

    def sample(self, model_output: np.ndarray) -> np.ndarray:
        # model_output is of shape (n_series * n_samples, output_chunk_length, n_components, n_quantiles)
        # sample uniformly between [0, 1] (for each batch example) and return the
        # linear interpolation between the fitted quantiles closest to the sampled value.
        k, n_times, n_components, n_quantiles = model_output.shape

        # obtain samples
        probs = self._rng.uniform(
            size=(
                k,
                n_times,
                n_components,
                1,
            )
        )

        # add dummy dim
        probas = np.expand_dims(probs, axis=-2)

        # tile and transpose
        p = np.tile(probas, (1, 1, 1, n_quantiles, 1)).transpose((0, 1, 2, 4, 3))

        # prepare quantiles
        tquantiles = np.array(self.quantiles).reshape((1, 1, 1, -1))

        # calculate index of the largest quantile smaller than the sampled value
        left_idx = np.sum(p > tquantiles, axis=-1)

        # obtain index of the smallest quantile larger than the sampled value
        right_idx = left_idx + 1

        # repeat the model output on the edges
        repeat_count = [1] * n_quantiles
        repeat_count[0] = 2
        repeat_count[-1] = 2
        repeat_count = np.array(repeat_count)
        shifted_output = np.repeat(model_output, repeat_count, axis=-1)

        # obtain model output values corresponding to the quantiles left and right of the sampled value
        left_value = np.take_along_axis(shifted_output, left_idx, axis=-1)
        right_value = np.take_along_axis(shifted_output, right_idx, axis=-1)

        # add 0 and 1 to quantiles
        ext_quantiles = [0.0] + self.quantiles + [1.0]
        expanded_q = np.tile(np.array(ext_quantiles), left_idx.shape)

        # calculate closest quantiles to the sampled value
        left_q = np.take_along_axis(expanded_q, left_idx, axis=-1)
        right_q = np.take_along_axis(expanded_q, right_idx, axis=-1)

        # linear interpolation
        weights = (probs - left_q) / (right_q - left_q)
        inter = left_value + weights * (right_value - left_value)

        # shape (n_series * n_samples, output_chunk_length, n_components * n_quantiles)
        return inter.squeeze(-1)

    def predict_likelihood_parameters(self, model_output: np.ndarray) -> np.ndarray:
        # shape (n_series * n_samples, output_chunk_length, n_components, n_quantiles)
        # quantiles on the last dimension, grouped by component
        k, n_times, n_components, n_quantiles = model_output.shape
        # last dim : [comp_1_q_1, ..., comp_1_q_n, ..., comp_n_q_1, ..., comp_n_q_n]
        # shape (n_series * n_samples, output_chunk_length, n_components * n_quantiles)
        return model_output.reshape(k, n_times, n_components * n_quantiles)

    def _estimator_predict(
        self,
        model,
        x: np.ndarray,
        **kwargs,
    ) -> np.ndarray:
        # `x` is of shape (n_series * n_samples, n_regression_features)
        k = x.shape[0]
        model_outputs = []
        for quantile, fitted in model._model_container.items():
            model.model = fitted
            # model output has shape (n_series * n_samples, output_chunk_length, n_components)
            model_output = fitted.predict(x, **kwargs).reshape(k, self._n_outputs, -1)
            model_outputs.append(model_output)
        model_outputs = np.stack(model_outputs, axis=-1)
        # shape (n_series * n_samples, output_chunk_length, n_components, n_quantiles)
        return model_outputs

    def _get_median_prediction(self, model_output: np.ndarray) -> np.ndarray:
        # shape (n_series * n_samples, output_chunk_length, n_components, n_quantiles)
        k, n_times, n_components, n_quantiles = model_output.shape
        # extract the median quantiles per component
        component_medians = slice(self._median_idx, None, n_quantiles)
        # shape (n_series * n_samples, output_chunk_length, n_components)
        return model_output[:, :, :, component_medians].reshape(
            k, n_times, n_components
        )


def _check_likelihood(likelihood: str, available_likelihoods: list[str]):
    """Check whether the likelihood is supported.

    Parameters
    ----------
    likelihood
        The likelihood name. Must be one of ('gaussian', 'poisson', 'quantile').
    available_likelihoods
        A list of supported likelihood names.
    """
    if likelihood not in available_likelihoods:
        raise_log(
            ValueError(
                f"Invalid `likelihood='{likelihood}'`. Must be one of {available_likelihoods}"
            ),
            logger=logger,
        )


def _get_likelihood(
    likelihood: Optional[str],
    n_outputs: int,
    random_state: Optional[int],
    quantiles: Optional[list[float]],
) -> Optional[SKLearnLikelihood]:
    """Get the `Likelihood` object for `RegressionModel`.

    Parameters
    ----------
    likelihood
        The likelihood name. Must be one of ('gaussian', 'poisson', 'quantile').
    n_outputs
        The number of predicted outputs per model call. `1` if `multi_models=False`, otherwise
        `output_chunk_length`.
    random_state
        Optionally, control the randomness of the sampling.
    quantiles
        Optionally, a list of quantiles. Only effective for `likelihood='quantile'`.
    """
    if likelihood is None:
        return None
    elif likelihood == "gaussian":
        return GaussianLikelihood(n_outputs=n_outputs, random_state=random_state)
    elif likelihood == "poisson":
        return PoissonLikelihood(n_outputs=n_outputs, random_state=random_state)
    elif likelihood == "quantile":
        return QuantileRegression(
            n_outputs=n_outputs, random_state=random_state, quantiles=quantiles
        )
    else:
        raise_log(
            ValueError(
                f"Invalid `likelihood='{likelihood}'`. Must be one of ('gaussian', 'poisson', 'quantile')"
            ),
            logger=logger,
        )



"""
Croston method
--------------
"""

import copy
from typing import Optional

from statsforecast.models import TSB as CrostonTSB
from statsforecast.models import CrostonClassic, CrostonOptimized, CrostonSBA

from darts.logging import get_logger, raise_log
from darts.models.forecasting.sf_model import StatsForecastModel

logger = get_logger(__name__)


class Croston(StatsForecastModel):
    def __init__(
        self,
        version: str = "classic",
        alpha_d: float = None,
        alpha_p: float = None,
        add_encoders: Optional[dict] = None,
        quantiles: Optional[list[float]] = None,
        random_state: Optional[int] = None,
        **kwargs,
    ):
        """Croston method as presented `in this paper <https://otexts.com/fpp3/counts.html>`_ and based on the
        `Statsforecasts package <https://github.com/Nixtla/statsforecast>`_.

        We refer to the StatsForecast documentation of
        `CrostonClassic <https://nixtlaverse.nixtla.io/statsforecast/src/core/models.html#crostonclassic>`_,
        `CrostonOptimized <https://nixtlaverse.nixtla.io/statsforecast/src/core/models.html#crostonoptimized>`_,
        `CrostonSBA <https://nixtlaverse.nixtla.io/statsforecast/src/core/models.html#crostonsba>`_, and
        `Teunter-Syntetos-Babai <https://nixtlaverse.nixtla.io/statsforecast/src/core/models.html#tsb>`_.

        In addition to univariate deterministic forecasting, it comes with additional support:

        - **Future covariates:** Use exogenous features to potentially improve predictive accuracy.
          Darts adds support by first regressing the series against the future covariates using a
          :class:`~darts.models.forecasting.linear_regression_model.LinearRegressionModel` model and then running the
          StatsForecast model on the in-sample residuals from this original regression. This approach was inspired by
          `this post of Stephan Kolassa <https://stats.stackexchange.com/q/220885>`_.

        - **Probabilstic / Conformal forecasting:** Probabilstic forecasting can be performed using conformal
          prediction. To activate it, simply set `prediction_intervals` at model creation. To generate probabilistic
          forecasts, you can set the following parameters when calling
          :meth:`~darts.models.forecasting.sf_model.StatsForecastModel.predict`:

          - Forecast quantile values directly by setting `predict_likelihood_parameters=True`.

          - Generate sampled forecasts from these quantiles by setting `num_samples >> 1`.

        - **Transferable series forecasting:** Apply the fitted model to a new input `series` at prediction time.
          Darts adds support by first fitting a copy of the model on the new series, and then using that model to
          generate the corresponding forecast.

        .. note::
            Future covariates are not supported when the input series contain missing values.

        .. note::
            The first model call might take more time than all subsequent calls as the model relies on Numba and jit
            compilation.

        Parameters
        ----------
        version
            - "classic" corresponds to classic Croston.
            - "optimized" corresponds to optimized classic Croston, which searches
              for the optimal ``alpha`` smoothing parameter and can take longer
              to run. Otherwise, a fixed value of ``alpha=0.1`` is used.
            - "sba" corresponds to the adjustment of the Croston method known as
              the Syntetos-Boylan Approximation [1]_.
            - "tsb" corresponds to the adjustment of the Croston method proposed by
              Teunter, Syntetos and Babai [2]_. In this case, `alpha_d` and `alpha_p` must
              be set.
        alpha_d
            For the "tsb" version, the alpha smoothing parameter to apply on demand.
        alpha_p
            For the "tsb" version, the alpha smoothing parameter to apply on probability.
        add_encoders
            A large number of future covariates can be automatically generated with `add_encoders`.
            This can be done by adding multiple pre-defined index encoders and/or custom user-made functions that
            will be used as index encoders. Additionally, a transformer such as Darts' :class:`Scaler` can be added to
            transform the generated covariates. This happens all under one hood and only needs to be specified at
            model creation.
            Read :meth:`SequentialEncoder <darts.dataprocessing.encoders.SequentialEncoder>` to find out more about
            ``add_encoders``. Default: ``None``. An example showing some of ``add_encoders`` features:

            .. highlight:: python
            .. code-block:: python

                def encode_year(idx):
                    return (idx.year - 1950) / 50

                add_encoders={
                    'cyclic': {'future': ['month']},
                    'datetime_attribute': {'future': ['hour', 'dayofweek']},
                    'position': {'future': ['relative']},
                    'custom': {'future': [encode_year]},
                    'transformer': Scaler(),
                    'tz': 'CET'
                }
            ..
        quantiles
            Optionally, produce quantile predictions at `quantiles` levels when performing probabilistic forecasting
            with `num_samples > 1` or `predict_likelihood_parameters=True`.
        random_state
            Control the randomness of probabilistic conformal forecasts (sample generation) across different runs.
        kwargs
            Keyword arguments for ``statsforecasts.models.Croston*``.

        References
        ----------
        .. [1] Aris A. Syntetos and John E. Boylan. The accuracy of intermittent demand estimates.
               International Journal of Forecasting, 21(2):303 – 314, 2005.
        .. [2] Ruud H. Teunter, Aris A. Syntetos, and M. Zied Babai.
               Intermittent demand: Linking forecasting to inventory obsolescence.
               European Journal of Operational Research, 214(3):606 – 615, 2011.

        Examples
        --------
        >>> from darts.datasets import AirPassengersDataset
        >>> from darts.models import Croston
        >>> from darts.utils.timeseries_generation import datetime_attribute_timeseries
        >>> series = AirPassengersDataset().load()
        >>> # optionally, use some future covariates; e.g. the value of the month encoded as a sine and cosine series
        >>> future_cov = datetime_attribute_timeseries(series, "month", cyclic=True, add_length=6)
        >>> # define Croston parameters
        >>> model = Croston(version="optimized")
        >>> model.fit(series, future_covariates=future_cov)
        >>> pred = model.predict(6, future_covariates=future_cov)
        >>> pred.values()
        array([[419.84565922],
               [424.06484452],
               [440.05509455],
               [463.53183473],
               [488.20449148],
               [507.46204636]])
        """
        if version.lower() not in ["classic", "optimized", "sba", "tsb"]:
            raise_log(
                ValueError(
                    'The provided "version" parameter must be set to "classic", "optimized", "sba" or "tsb".'
                ),
                logger=logger,
            )

        kwargs = copy.deepcopy(kwargs)
        if version == "classic":
            model_cls = CrostonClassic
        elif version == "optimized":
            model_cls = CrostonOptimized
        elif version == "sba":
            model_cls = CrostonSBA
        else:
            if alpha_d is None or alpha_p is None:
                raise_log(
                    ValueError(
                        'alpha_d and alpha_p must be specified when using "tsb".'
                    ),
                    logger=logger,
                )
            kwargs["alpha_d"] = alpha_d
            kwargs["alpha_p"] = alpha_p
            model_cls = CrostonTSB

        model = model_cls(**kwargs)
        super().__init__(
            model=model,
            quantiles=quantiles,
            add_encoders=add_encoders,
            random_state=random_state,
        )



"""
Dynamic Time Warping (DTW)
--------------------------
"""

from darts.dataprocessing.dtw.cost_matrix import CostMatrix
from darts.dataprocessing.dtw.dtw import DTWAlignment, dtw
from darts.dataprocessing.dtw.window import (
    CRWindow,
    Itakura,
    NoWindow,
    SakoeChiba,
    Window,
)

__all__ = [
    "CostMatrix",
    "DTWAlignment",
    "dtw",
    "CRWindow",
    "Itakura",
    "NoWindow",
    "SakoeChiba",
    "Window",
]



"""
Filtering Anomaly Model
-----------------------

A `FilteringAnomalyModel` wraps around a Darts filtering model and one or
several anomaly scorer(s) to compute anomaly scores
by comparing how actuals deviate from the model's predictions (filtered series).
"""

import sys
from collections.abc import Sequence
from typing import Literal, Optional, Union

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

from darts.ad.anomaly_model.anomaly_model import AnomalyModel
from darts.ad.scorers.scorers import AnomalyScorer
from darts.logging import get_logger, raise_log
from darts.models.filtering.filtering_model import FilteringModel
from darts.timeseries import TimeSeries

logger = get_logger(__name__)


class FilteringAnomalyModel(AnomalyModel):
    def __init__(
        self,
        model: FilteringModel,
        scorer: Union[AnomalyScorer, Sequence[AnomalyScorer]],
    ):
        """Filtering-based Anomaly Detection Model

        The filtering model may or may not be already fitted. The underlying assumption is that this model
        should be able to adequately filter the series in the absence of anomalies. For this reason,
        it is recommended to either provide a model that has already been fitted and evaluated to work
        appropriately on a series without anomalies, or to ensure that a simple call to the :func:`fit()`
        function of the model will be sufficient to train it to satisfactory performance on series without anomalies.

        Calling :func:`fit()` on the anomaly model will fit the underlying filtering model only
        if `allow_model_training` is set to `True` upon calling `fit()`.
        In addition, calling :func:`fit()` will also fit the fittable scorers, if any.

        Parameters
        ----------
        model
            A Darts `FilteringModel` used to filter the actual time series.
        scorer
            One or multiple scorer(s) used to compare the actual and predicted time series in order to obtain an
            anomaly score `TimeSeries`. If a list of scorers,
            :meth:`~darts.ad.anomaly_model.filtering_am.FilteringAnomalyModel.score` will output anomaly scores for
            each scorer.
        """
        if not isinstance(model, FilteringModel):
            raise_log(
                ValueError("`model` must be a Darts `FilteringModel`."),
                logger=logger,
            )
        super().__init__(model=model, scorer=scorer)

    def fit(
        self,
        series: Union[TimeSeries, Sequence[TimeSeries]],
        allow_model_training: bool = False,
        **filter_fit_kwargs,
    ) -> Self:
        """Fit the underlying filtering model (if applicable) and the fittable scorers, if any.

        Train the filter (if not already fitted and `allow_model_training` is `True`) and the fittable scorer(s) on the
        given time series.

        The filter model will be applied to the given series, and the results will be used
        to train the scorer(s).

        Parameters
        ----------
        series
            The (sequence of) series to train on (generally assumed to be anomaly-free).
        allow_model_training
            Whether the filtering model should be fitted on the given series. If `False`, the model must already be
            fitted.
        **filter_fit_kwargs
            Additional parameters passed to the filtering model's `fit()` method.

        Returns
        -------
        self
            Fitted model.
        """
        return super().fit(
            series=series,
            allow_model_training=allow_model_training,
            **filter_fit_kwargs,
        )

    def score(
        self,
        series: Union[TimeSeries, Sequence[TimeSeries]],
        return_model_prediction: bool = False,
        **filter_kwargs,
    ) -> Union[TimeSeries, Sequence[TimeSeries], Sequence[Sequence[TimeSeries]]]:
        """Compute the anomaly score(s) for the given (sequence of) series.

        Predicts the given target time series with the filtering model, and applies the scorer(s)
        to compare the predicted (filtered) series and the provided series.

        Parameters
        ----------
        series
            The (sequence of) series to score on.
        return_model_prediction
            Whether to return the filtering model prediction along with the anomaly scores.
        **filter_kwargs
            Additional parameters passed to the filtering model's `filter()` method.

        Returns
        -------
        TimeSeries
            A single `TimeSeries` for a single `series` with a single anomaly scorers.
        Sequence[TimeSeries]
            A sequence of `TimeSeries` for:

            - a single `series` with multiple anomaly scorers.
            - a sequence of `series` with a single anomaly scorer.
        Sequence[Sequence[TimeSeries]]
            A sequence of sequences of `TimeSeries` for a sequence of `series` and multiple anomaly scorers.
            The outer sequence is over the series, and inner sequence is over the scorers.
        """
        return super().score(
            series=series,
            return_model_prediction=return_model_prediction,
            **filter_kwargs,
        )

    def predict_series(
        self, series: Sequence[TimeSeries], **kwargs
    ) -> Sequence[TimeSeries]:
        """Filters the given sequence of target time series with the filtering model.

        Parameters
        ----------
        series
            The sequence of series to filter.
        **kwargs
            Additional parameters passed to the filtering model's `filter()` method.
        """
        return [self.model.filter(s, **kwargs) for s in series]

    def eval_metric(
        self,
        anomalies: Union[TimeSeries, Sequence[TimeSeries]],
        series: Union[TimeSeries, Sequence[TimeSeries]],
        metric: Literal["AUC_ROC", "AUC_PR"] = "AUC_ROC",
        **filter_kwargs,
    ) -> Union[
        dict[str, float],
        dict[str, Sequence[float]],
        Sequence[dict[str, float]],
        Sequence[dict[str, Sequence[float]]],
    ]:
        """Compute a metric for the anomaly scores computed by the model.

        Predicts the `series` with the filtering model, and applies the scorer(s) on the filtered time series
        and the given target time series. Returns the score(s) of an agnostic threshold metric, based on the anomaly
        score given by the scorer(s).

        Parameters
        ----------
        anomalies
            The (sequence of) ground truth binary anomaly series (`1` if it is an anomaly and `0` if not).
        series
            The (sequence of) series to predict anomalies on.
        metric
            The name of the metric function to use. Must be one of "AUC_ROC" (Area Under the
            Receiver Operating Characteristic Curve) and "AUC_PR" (Average Precision from scores).
            Default: "AUC_ROC".
        **filter_kwargs
            Additional parameters passed to the filtering model's `filter()` method.

        Returns
        -------
        Dict[str, float]
            A dictionary with the resulting metrics for single univariate `series`, with keys representing the
            anomaly scorer(s), and values representing the metric values.
        Dict[str, Sequence[float]]
            Same as for `Dict[str, float]` but for multivariate `series`, and anomaly scorers that treat series
            components/columns independently (by nature of the scorer or if `component_wise=True`).
        Sequence[Dict[str, float]]
            Same as for `Dict[str, float]` but for a sequence of univariate series.
        Sequence[Dict[str, Sequence[float]]]
            Same as for `Dict[str, float]` but for a sequence of multivariate series.
        """
        return super().eval_metric(
            anomalies=anomalies,
            series=series,
            metric=metric,
            **filter_kwargs,
        )

    def show_anomalies(
        self,
        series: TimeSeries,
        anomalies: TimeSeries = None,
        names_of_scorers: Union[str, Sequence[str]] = None,
        title: str = None,
        metric: Optional[Literal["AUC_ROC", "AUC_PR"]] = None,
        **score_kwargs,
    ):
        """Plot the results of the anomaly model.

        Computes the score on the given series input and shows the different anomaly scores with respect to time.

        The plot will be composed of the following:

        - the series itself with the output of the forecasting model.
        - the anomaly score for each scorer. The scorers with different windows will be separated.
        - the actual anomalies, if given.

        It is possible to:

        - add a title to the figure with the parameter `title`
        - give personalized names for the scorers with `names_of_scorers`
        - show the results of a metric for each anomaly score (AUC_ROC or AUC_PR), if the actual anomalies are provided.

        Parameters
        ----------
        series
            The series to visualize anomalies from.
        anomalies
            The ground truth of the anomalies (1 if it is an anomaly and 0 if not).
        names_of_scorers
            Name of the scores. Must be a list of length equal to the number of scorers in the anomaly_model.
        title
            Title of the figure.
        metric
            Optionally, the name of the metric function to use. Must be one of "AUC_ROC" (Area Under the
            Receiver Operating Characteristic Curve) and "AUC_PR" (Average Precision from scores).
            Default: "AUC_ROC".
        score_kwargs
            parameters for the `score()` method.
        """
        return super().show_anomalies(
            series=series,
            anomalies=anomalies,
            predict_kwargs=None,
            names_of_scorers=names_of_scorers,
            title=title,
            metric=metric,
            **score_kwargs,
        )

    def _fit_core(
        self,
        series: Sequence[TimeSeries],
        allow_model_training: bool,
        **model_fit_kwargs,
    ):
        """Fit the filters (if applicable) and scorers."""
        # TODO: add support for covariates (see eg. Kalman Filter)
        if allow_model_training and hasattr(self.model, "fit"):
            # TODO: check if filter is already fitted (for now fit it regardless -> only Kalman)
            if len(series) > 1:
                raise_log(
                    ValueError(
                        f"Filter model {self.model.__class__.__name__} can only be fitted "
                        f"on a single time series, but multiple are provided."
                    ),
                    logger=logger,
                )
            self.model.fit(series[0], **model_fit_kwargs)
        else:
            # TODO: check if Kalman is fitted or not
            # if not raise error "fit filter before, or set `allow_model_training` to TRUE"
            pass

        if self.scorers_are_trainable:
            pred = self.predict_series(series)
            # fit the scorers
            self._fit_scorers(series, pred)



from collections.abc import Mapping, Sequence
from typing import Any, Union

import numpy as np

from darts import TimeSeries
from darts.dataprocessing.transformers.invertible_data_transformer import (
    InvertibleDataTransformer,
)
from darts.utils.timeseries_generation import constant_timeseries


class TestInvertibleDataTransformer:
    class DataTransformerMock(InvertibleDataTransformer):
        def __init__(
            self,
            scale: float,
            translation: float,
            stack_samples: bool = False,
            mask_components: bool = True,
            parallel_params: Union[bool, Sequence[str]] = False,
        ):
            """
            Applies the (invertible) transform `transformed_series = scale * series + translation`.

            Parameters
            ----------
            scale
                Scale coefficient of transform.
            translation
                Translational constant of transform.
            stack_samples
                Whether to call `stack_samples` inside of `ts_transform`.
            mask_components
                Whether to automatically apply any provided `component_mask` key word arguments. See
                `BaseDataTransformer` docstring for further details.
            parallel_params
                Specifies which parameters should vary between different parallel jobs, supposing that
                multiple time series are given to `ts_transform`. See `BaseDataTransformer` docstring
                for further details.

            """
            self._scale = scale
            self._translation = translation
            self._stack_samples = stack_samples
            self._mask_components = mask_components
            super().__init__(
                name="DataTransformerMock",
                mask_components=mask_components,
                parallel_params=parallel_params,
            )

        @staticmethod
        def ts_transform(
            series: TimeSeries, params: Mapping[str, Any], **kwargs
        ) -> TimeSeries:
            """
            Implements the transform `scale * series + translation`.

            If `component_mask` is in `kwargs`, this is manually applied and unapplied. If
            `_stack_samples = True` in `params['fixed']`, then `stack_samples` and `unstack_samples`
            all used when computing this transformation.

            """
            fixed_params = params["fixed"]
            stack_samples = fixed_params["_stack_samples"]
            mask_components = fixed_params["_mask_components"]
            scale, translation = fixed_params["_scale"], fixed_params["_translation"]

            # Ensure `component_mask` not passed via `kwargs` if specified `mask_components=True`
            # when transform created
            if mask_components:
                assert "component_mask" not in kwargs

            # Ensure manual masking only performed when `mask_components = False`
            # when transform constructed:
            if not mask_components and ("component_mask" in kwargs):
                vals = TestInvertibleDataTransformer.DataTransformerMock.apply_component_mask(
                    series, kwargs["component_mask"], return_ts=False
                )
            else:
                vals = series.all_values()

            if stack_samples:
                vals = TestInvertibleDataTransformer.DataTransformerMock.stack_samples(
                    vals
                )
            vals = scale * vals + translation
            if stack_samples:
                vals = (
                    TestInvertibleDataTransformer.DataTransformerMock.unstack_samples(
                        vals, series=series
                    )
                )

            if not mask_components and ("component_mask" in kwargs):
                vals = TestInvertibleDataTransformer.DataTransformerMock.unapply_component_mask(
                    series, vals, kwargs["component_mask"]
                )

            return series.with_values(vals)

        @staticmethod
        def ts_inverse_transform(
            series: TimeSeries, params: Mapping[str, Any], **kwargs
        ) -> TimeSeries:
            """
            Implements the inverse transform `(series - translation) / scale`.

            If `component_mask` is in `kwargs`, this is manually applied and unapplied. If
            `_stack_samples = True` in `params['fixed']`, then `stack_samples` and `unstack_samples`
            all used when computing this transformation.

            """
            fixed_params = params["fixed"]
            stack_samples = fixed_params["_stack_samples"]
            mask_components = fixed_params["_mask_components"]
            scale, translation = fixed_params["_scale"], fixed_params["_translation"]

            # Ensure `component_mask` not passed via `kwargs` if specified `mask_components=True`
            # when transform created
            if mask_components:
                assert "component_mask" not in kwargs

            # Ensure manual masking only performed when `mask_components = False`
            # when transform constructed:
            if not mask_components and ("component_mask" in kwargs):
                vals = TestInvertibleDataTransformer.DataTransformerMock.apply_component_mask(
                    series, kwargs["component_mask"], return_ts=False
                )
            else:
                vals = series.all_values()

            if stack_samples:
                vals = TestInvertibleDataTransformer.DataTransformerMock.stack_samples(
                    vals
                )
            vals = (vals - translation) / scale
            if stack_samples:
                vals = (
                    TestInvertibleDataTransformer.DataTransformerMock.unstack_samples(
                        vals, series=series
                    )
                )

            if not mask_components and ("component_mask" in kwargs):
                vals = TestInvertibleDataTransformer.DataTransformerMock.unapply_component_mask(
                    series, vals, kwargs["component_mask"]
                )

            return series.with_values(vals)

    def test_input_transformed_single_series(self):
        """
        Tests for correct (inverse) transformation of single series.
        """
        test_input = constant_timeseries(value=1, length=10)

        mock = self.DataTransformerMock(scale=2, translation=10)

        transformed = mock.transform(test_input)

        # 2 * 1 + 10 = 12
        expected = constant_timeseries(value=12, length=10)
        assert transformed == expected

        # Should get input back:
        assert mock.inverse_transform(transformed) == test_input

    def test_input_transformed_multiple_series(self):
        """
        Tests for correct transformation of multiple series when
        different param values are used for different parallel
        jobs (i.e. test that `parallel_params` argument is treated
        correctly). Also tests that transformer correctly handles
        being provided with fewer input series than fixed parameter
        value sets.
        """
        test_input_1 = constant_timeseries(value=1, length=10)
        test_input_2 = constant_timeseries(value=2, length=11)

        # Don't have different params for different jobs:
        mock = self.DataTransformerMock(scale=2, translation=10, parallel_params=False)
        (transformed_1, transformed_2) = mock.transform((test_input_1, test_input_2))
        # 2 * 1 + 10 = 12
        assert transformed_1 == constant_timeseries(value=12, length=10)
        # 2 * 2 + 10 = 14
        assert transformed_2 == constant_timeseries(value=14, length=11)
        # Should get input back:
        inv_1, inv_2 = mock.inverse_transform((transformed_1, transformed_2))
        assert inv_1 == test_input_1
        assert inv_2 == test_input_2

        # If only one timeseries provided, should apply parameters defined for
        # for the first to that series:
        transformed_1 = mock.transform(test_input_1)
        assert transformed_1 == constant_timeseries(value=12, length=10)
        inv_1 = mock.inverse_transform(transformed_1)
        assert inv_1 == test_input_1

        # Have different `scale` param for different jobs:
        mock = self.DataTransformerMock(
            scale=(2, 3), translation=10, parallel_params=["_scale"]
        )
        (transformed_1, transformed_2) = mock.transform((test_input_1, test_input_2))
        # 2 * 1 + 10 = 12
        assert transformed_1 == constant_timeseries(value=12, length=10)
        # 3 * 2 + 10 = 16
        assert transformed_2 == constant_timeseries(value=16, length=11)
        # Should get input back:
        inv_1, inv_2 = mock.inverse_transform((transformed_1, transformed_2))
        assert inv_1 == test_input_1
        assert inv_2 == test_input_2

        # Have different `scale`, `translation`, and `stack_samples` params for different jobs:
        mock = self.DataTransformerMock(
            scale=(2, 3),
            translation=(10, 11),
            stack_samples=(False, True),
            mask_components=(False, False),
            parallel_params=True,
        )
        (transformed_1, transformed_2) = mock.transform((test_input_1, test_input_2))
        # 2 * 1 + 10 = 12
        assert transformed_1 == constant_timeseries(value=12, length=10)
        # 3 * 2 + 11 = 17
        assert transformed_2 == constant_timeseries(value=17, length=11)
        # Should get input back:
        inv_1, inv_2 = mock.inverse_transform((transformed_1, transformed_2))
        assert inv_1 == test_input_1
        assert inv_2 == test_input_2
        # If only one timeseries provided, should apply parameters defined for
        # for the first to that series:
        transformed_1 = mock.transform(test_input_1)
        assert transformed_1 == constant_timeseries(value=12, length=10)
        inv_1 = mock.inverse_transform(transformed_1)
        assert inv_1 == test_input_1

        # Specify three sets of fixed params, but pass only one or two series as inputs
        # to `transform`/`inverse_transform`; transformer should apply `i`th set of fixed
        # params to the `i`th input passed to `transform`
        mock = self.DataTransformerMock(
            scale=(2, 3, 4),
            translation=(10, 11, 12),
            stack_samples=(False, True, False),
            mask_components=(False, False, False),
            parallel_params=True,
        )
        # If single series provided to transformer with three sets of
        # fixed params, should transform/inverse transform using the first set of fixed
        # parameters:
        transformed_1 = mock.transform(test_input_1)
        assert transformed_1 == constant_timeseries(value=12, length=10)
        inv_1 = mock.inverse_transform(transformed_1)
        assert inv_1 == test_input_1
        # If two series provided to transformer with three sets of
        # fixed params, should transform/inverse transform using the first and
        # second set of fixed parameters:
        transformed_1, transformed_2 = mock.transform((test_input_1, test_input_2))
        assert transformed_1 == constant_timeseries(value=12, length=10)
        assert transformed_2 == constant_timeseries(value=17, length=11)
        inv_1, inv_2 = mock.inverse_transform((transformed_1, transformed_2))
        assert inv_1 == test_input_1
        assert inv_2 == test_input_2

    def test_input_transformed_list_of_lists_of_series(self):
        """
        Tests for correct transformation of multiple series when
        different param values are used for different parallel
        jobs (i.e. test that `parallel_params` argument is treated
        correctly). Also tests that transformer correctly handles
        being provided with fewer input series than fixed parameter
        value sets.
        """
        test_input_1 = constant_timeseries(value=1, length=10)
        test_input_2 = constant_timeseries(value=2, length=11)

        # Don't have different params for different jobs:
        mock = self.DataTransformerMock(scale=2, translation=10, parallel_params=False)
        (transformed_1, transformed_2) = mock.transform((test_input_1, test_input_2))
        # 2 * 1 + 10 = 12
        assert transformed_1 == constant_timeseries(value=12, length=10)
        # 2 * 2 + 10 = 14
        assert transformed_2 == constant_timeseries(value=14, length=11)

        # list of lists of series must get input back
        inv = mock.inverse_transform([[transformed_1], [transformed_2]])
        assert len(inv) == 2
        assert all(
            isinstance(series_list, list) and len(series_list) == 1
            for series_list in inv
        )
        assert all(
            isinstance(series, TimeSeries)
            for series_list in inv
            for series in series_list
        )
        assert inv[0][0] == test_input_1
        assert inv[1][0] == test_input_2

        # one list of lists of is longer than others, must get input back
        inv = mock.inverse_transform([[transformed_1, transformed_1], [transformed_2]])
        assert len(inv) == 2
        assert len(inv[0]) == 2 and len(inv[1]) == 1
        assert all(isinstance(series_list, list) for series_list in inv)
        assert all(
            isinstance(series, TimeSeries)
            for series_list in inv
            for series in series_list
        )
        assert inv[0][0] == test_input_1
        assert inv[0][1] == test_input_1
        assert inv[1][0] == test_input_2

        # different types of Sequences, must get input back
        inv = mock.inverse_transform(((transformed_1, transformed_1), (transformed_2,)))
        assert len(inv) == 2
        assert len(inv[0]) == 2 and len(inv[1]) == 1
        assert all(isinstance(series_list, list) for series_list in inv)
        assert all(
            isinstance(series, TimeSeries)
            for series_list in inv
            for series in series_list
        )
        assert inv[0][0] == test_input_1
        assert inv[0][1] == test_input_1
        assert inv[1][0] == test_input_2

        # one list of lists is empty, returns empty list as well
        inv = mock.inverse_transform([[], [transformed_2, transformed_2]])
        assert len(inv) == 2
        assert len(inv[0]) == 0 and len(inv[1]) == 2
        assert all(isinstance(series_list, list) for series_list in inv)
        assert all(isinstance(series, TimeSeries) for series in inv[1])
        assert inv[1][0] == test_input_2
        assert inv[1][1] == test_input_2

        # more list of lists than used during transform works
        inv = mock.inverse_transform([
            [transformed_1],
            [transformed_2],
            [transformed_2],
        ])
        assert len(inv) == 3
        assert all(
            isinstance(series_list, list) and len(series_list) == 1
            for series_list in inv
        )
        assert all(
            isinstance(series, TimeSeries)
            for series_list in inv
            for series in series_list
        )
        assert inv[0][0] == test_input_1
        assert inv[1][0] == test_input_2
        assert inv[2][0] == test_input_2

    def test_input_transformed_multiple_samples(self):
        """
        Tests that `stack_samples` and `unstack_samples` correctly
        implemented when considering multi-sample timeseries.
        """
        test_input = constant_timeseries(value=1, length=10)
        test_input = test_input.concatenate(
            constant_timeseries(value=2, length=10), axis="sample"
        )

        mock = self.DataTransformerMock(scale=2, translation=10, stack_samples=True)
        transformed = mock.transform(test_input)

        # 2 * 1 + 10 = 12
        expected = constant_timeseries(value=12, length=10)
        # 2 * 2 + 10 = 14
        expected = expected.concatenate(
            constant_timeseries(value=14, length=10), axis="sample"
        )
        assert transformed == expected
        # Should get input back:
        inv = mock.inverse_transform(transformed)
        assert inv == test_input

    def test_input_transformed_masking(self):
        """
        Tests that automatic component masking is correctly implemented,
        and that manual component masking is also handled correctly
        through `kwargs` + `apply_component_mask`/`unapply_component_mask`
        methods.
        """
        test_input = TimeSeries.from_values(np.ones((4, 3, 5)))
        mask = np.array([True, False, True])
        # Second component should be untransformed:
        scale = 2
        translation = 10
        expected = np.stack(
            [12 * np.ones((4, 5)), np.ones((4, 5)), 12 * np.ones((4, 5))], axis=1
        )
        expected = TimeSeries.from_values(expected)

        # Automatically apply component mask:
        mock = self.DataTransformerMock(
            scale=scale, translation=translation, mask_components=True
        )
        transformed = mock.transform(test_input, component_mask=mask)
        assert transformed == expected
        # Should get input back:
        inv = mock.inverse_transform(transformed, component_mask=mask)
        assert inv == test_input

        # Manually apply component mask:
        mock = self.DataTransformerMock(scale=2, translation=10, mask_components=False)
        transformed = mock.transform(test_input, component_mask=mask)
        assert transformed == expected
        # Should get input back:
        inv = mock.inverse_transform(transformed, component_mask=mask)
        assert inv == test_input



from itertools import combinations

from darts.utils.likelihood_models.sklearn import (
    GaussianLikelihood,
    PoissonLikelihood,
    QuantileRegression,
)

# equality between likelihoods is only dependent on the main distribution parameters
likelihood_models_equal = {
    "quantile": [
        QuantileRegression(n_outputs=1, quantiles=[0.1, 0.5, 0.9], random_state=1),
        QuantileRegression(n_outputs=12, quantiles=[0.1, 0.5, 0.9], random_state=2),
    ],
    "gaussian": [
        GaussianLikelihood(n_outputs=1, random_state=1),
        GaussianLikelihood(n_outputs=12, random_state=2),
    ],
    "poisson": [
        PoissonLikelihood(n_outputs=1, random_state=1),
        PoissonLikelihood(n_outputs=12, random_state=2),
    ],
}

likelihood_models_unequal = {
    "quantile": [
        QuantileRegression(n_outputs=1, quantiles=[0.2, 0.5, 0.8], random_state=1),
        QuantileRegression(n_outputs=1, quantiles=[0.1, 0.5, 0.9], random_state=1),
    ],
}


class TestSKLearnLikelihoodModel:
    def test_intra_class_equality(self):
        for _, model_pair in likelihood_models_equal.items():
            assert model_pair[0] == model_pair[0]
            assert model_pair[1] == model_pair[1]
            assert model_pair[0] == model_pair[1]

        for _, model_pair in likelihood_models_unequal.items():
            assert model_pair[0] == model_pair[0]
            assert model_pair[1] == model_pair[1]
            assert model_pair[0] != model_pair[1]

    def test_inter_class_equality(self):
        model_combinations = combinations(likelihood_models_equal.keys(), 2)
        for first_model_name, second_model_name in model_combinations:
            assert (
                likelihood_models_equal[first_model_name][0]
                != likelihood_models_equal[second_model_name][0]
            )
